#!/usr/bin/perl -w
#
# fourX.pl v1.0
# Last modified October 3, 2016 14:09
# (c) Olaf R.P. Bininda-Emonds
#
# Purpose: Identifies species within a aligned DNA matrix according to the 4x
#          rule of Birky et al. (2005. Hydrobiologia 546: 29-45)
#
# Input:
#	1) One or more sequence data files (recognized formats: fasta, nexus,
#	   phylip, and Se-Al)
#	2) A nexus- or phylip-formatted treefile containing a description (with
#      branch lengths) of the optimal tree (optional; if not provided, the tree
#      will be estimated using RAxML)
#	3) Results of a ModelTEST model-fitting procedure as obtained using
#	   autoMT.pl and presented in the file autoMTresults.txt (optional; if not
#	   provided, the optimal model will be determined using autoMT.pl)
#
# Requires:
#	1) autoMT.pl + PHYML in the $PATH to determine the optimal model of
#      evolution
#	2) PAUP* in the $PATH to calculate corrected and uncorrected distances
#	3) RAxML in the $PATH to determine the optimal ML tree (optional)
#
# Usage: fourX.pl -d<filename> [-a<name>] [-b<number>] [-i<f|n|p|s>]
#		 [-k<number>] [K] [-N<number>] [-o<outgroup(s)>] [-p<name>] [-P<name>]
#		 [-r<name>] [-s<ML|NJ>] [-t<filename>] [-z] [-v] [-h]
#   options: -d<filename> = file containing raw sequence information
#			 -a<name> = name (including path) of autoMT.pl perl script in $PATH
#						(default = ~/bin/autoMT_working.pl)
#			 -b<number> = number of bootstrap replicates for optimality analyses
#						  (default = 1000)
#            -i<f|n|p|s> = format of sequence file (fasta (f), nexus (n),
#						  phylip (p), or Se-Al (s)); default = autodetect
#            -k<number> = minimum k / theta ratio to define a species
#						  (default = 4)
#            -K = force K2P + G model
#			 -N<number> = number of processors available for parallel processing
#						  (default = 1 (off))
#            -o<outgroup(s)> = name(s) of outgroup(s) for optimality analyses,
#							   with multiple names separated by a comma
#			 -p<name> = name (including path) of PhyML 3.x program in $PATH
#						(default = ~/bin/phyml)
#			 -P<name> = name (including path) of PAUP* program in $PATH
#						(default = paup64)
#			 -r<name> = name (including path) of RAxML program in $PATH
#						(default = raxmlHPC)
#			 -s<ML|NJ> = derive final tree using either NJ (with PAUP*) or ML
#						 (with RAxML; default)
#			 -t<filename> = file containing optimal tree (with branch lengths)
#							for sequence data in -d
#			 -z = allow negative branch lengths for NJ analyses
#            -v = verbose output
#            -h = print this message and quit

use strict;
use POSIX qw(floor ceil);
use File::Basename;
use Config;
use English;

# Set user-input defaults and associated parameters
	# I/O variables
		my $dataFile;
		my $treeFile;
		my $matrixFile;
		my $groupFile;
		my $searchStrategy = "ML";	# Other option in NJ
		my $baseName = "fourX";
		my $forceK2P = 0;
		my $setZero = "prohibit";
		
	# Program variables
		my $raxml = "~/bin/raxmlHPC";
		my $paup = "~/bin/paup64";
		my $phyml = "~/bin/phyml";
			my $numBoot = 1000;
			my $outGroups;
			my $numThreads = 1;
		my $autoMT = "~/bin/autoMT_working.pl";
		my $numParallel = 1;

	# BootStrip parameters
		# I/O parameters
			my $rootWord = "treeFile";
				my $optFile = $rootWord . ".tre";
				my $outFile = $rootWord . "_bootstrip.tre";
				my $logFile = "bootstrap.log";
				my $bootFile = "bootTrees.tre";
			my $paupFlag = 0;
	
			# Printing parameters
				my $decimalPlace = 3;
				my $postProb = 0;
				my $labelType = "branch";
				my $keepLengths = 0;
				my $averagePrint = 0;

		# dotPlotter parameters
			my ($nullDot, $fullDot);
			my (%taxonID, %plotCount);
	
		# Bootstrap parameters
			my %bootFreq;
			my $minFreq;
			my $tipValue = 100;
			my $burninLimit = 0;
				
	# 4X variables
		my $kTheta = 4;
		my (%aveP, %kTheta, %kThetaStatus);
		my %namePresent;

	# pairwiseReader variables
		my $printForm = "column";	# Options are column and matrix
		my $diagPrint = "below";	# Options are above, below, and both (latter for matrix only)
	
		my ($aboveFlag, $belowFlag) = (0, 0);
		my ($rowTaxon, $columnTaxon) = (0, 0);
		my ($startColumn, $endColumn);
		my (%aboveDiag, %belowDiag);
		
		my %taxonName;
		my $numTaxa = 0;
		my $ignoreUndefined = 1;
		my $paupVersion;
		
	# seqRead variables
		my $inputType;	# Options are fasta, nexus, phylip, and Se-Al
#			my ($header, $tempAcc, $tempName, $tempSeq, $fastaAcc);
#			my $owner = 0;
#			my $sealDelFlag = 0;
			my $globalGenCode = 1;
#		my ($inputFile, @fileList);
		my $seqType = "DNA";
		my (@accNum, %nameLabel, %sequence, %geneticCode, %seqLength, %binomLabel);
		my (@charsetList, %charsetStart, %charsetEnd);
		my (%deletedSeq, %finalSeq);
		my $seqCount;
#		my $delCount;
		my $ntax;
		my $gapChange = 1;
		my $skipDuplicates = 1;

	# Translation tables and ambiguity codes
#		my %transTable = ('1' => 'standard',
#						  '2' => 'vertebrate mitochondrial',
#						  '3' => 'yeast mitochondrial',
#						  '4' => 'mold, protozoan and colenterate mitochondrial and mycoplasam/spiroplasma',
#						  '5' => 'invertebrate mitochondrial',
#						  '6' => 'ciliate, dasycladacean and hexamita nuclear',
#						  '9' => 'echinoderm mitochondrial',
#						  '10' => 'euplotid nuclear',
#						  '11' => 'bacterial and plant plastid',
#						  '12' => 'alternative yeast nuclear',
#						  '13' => 'ascidian mitochondrial',
#						  '14' => 'alternative flatworm mitochondrial',
#						  '15' => 'Blepharisma nuclear',
#						  '16' => 'chlorophycean mitochondrial',
#						  '21' => 'trematode mitochondrial',
#						  '22' => 'Scenedesmus obliquus mitochondrial',
#						  '23' => 'Thraustochytrium mitochondrial');
		my %gb2seal = ('1' => '0',
					   '2' => '1',
					   '3' => '2',
					   '4' => '3',
					   '4' => '4',
					   '5' => '5',
					   '6' => '6',
					   '9' => '7',
					   '10' => '8',
					   '11' => '9',
					   '12' => '10',
					   '13' => '11',
					   '14' => '12',
					   '15' => '13',
					   '16' => '1',
					   '21' => '1',
					   '22' => '1',
					   '23' => '1');
		my %seal2gb = ('0' => '1',
					   '1' => '2',
					   '2' => '3',
					   '3' => '4',
					   '4' => '4',
					   '5' => '5',
					   '6' => '6',
					   '7' => '9',
					   '8' => '10',
					   '9' => '11',
					   '10' => '12',
					   '11' => '13',
					   '12' => '14',
					   '13' => '15');

	# seqPrint variables
		my $maxLength = 0;
		my $fastaPrint = 0;
			my $fastaOut;
		my $megaPrint = 0;
			my $megaOut;
		my $nexusPrint = 0;
			my $nexusOut;
		my ($phylipTradPrint, $phylipExtPrint) = (0, 0);
			my $phylipOut;
		my $sealPrint = 0;
			my $sealOut;
		my $stockholmPrint = 0;
			my $stockholmOut;
		my $outFormat = "sequential";
			my $interleaveLength = 80;
			my $fastaLength = 80;
		my $accPrint = 0;

	# treeRead variables
		my (%treeName, %treeDescription, %rootingStatus);
		my $unrootedFlag = 0;
		my $defaultRooting = "[&U]";	# Default is for trees to be unrooted
		my $supportFlag = 0;
		my @allTaxa;
		my $block = 1;
			my $treeBlock = 0;
		my $treeNum = 0;
		my $keepComments = 0;

	# parseTree parameters
		my (%ancCladeHeight, %dauCladeHeight);
		
		my @nodeList;
		my (%ParentClade, %daughterClade);
		my (%CladeDescrip, %open, %start, %stop);
		my (%geneNodeMember, %taxonClade, @GTtaxa, %cladeSize);
		my (%CladeSupport, %nodeDepth, %cladeDepth, %cladeHeight);
			# Set variables for root
#				$nodeDepth{GTnode0} = 0;
#				$cladeDepth{GTnode0} = 0;
#				$CladeSupport{GTnode1} = 0;	# Default; will be overriden by any real value		
#		my (%gene2ST);
		my $taxonFlag;
		my %polyClade;
		my ($branchFlag, $labelFlag) = (0, 0);
			my $removeBL = 0;
		my $branchType = "length";
		my %nodeLabel;
		my $parseError = 0;
		my $bushCheck = 0;

	# Miscellaneous variables
		my @errorList;
		my $quiet = 0;
		my $verbose = 0;
		my $debug = 0;
		my $perlScript = "fourX.pl";
			my $callString = join(' ', $perlScript, @ARGV);
		my $version = "1.0";
		my $usage = "Usage: $perlScript -d<filename> [-a<name>] [-b<number>] [-i<f|n|p|s>]\n";
			$usage .= "       [-k<number>] [-K] [-N<number>] [-o<outgroup(s)>] [-p<name>] [-P<name>]\n";
			$usage .= "       [-r<name>] [-s<ML|NJ>] [-t<filename>] [-z] [-v] [-h]";
		my $options = "Options: -d<filename> = file containing raw sequence information\n";
			$options .= "         -a<name> = name (including path) of autoMT.pl Perl script in \$PATH\n";
			$options .= "                    (default = $autoMT)\n";
			$options .= "         -b<number> = number of bootstrap replicates for optimality analyses\n";
			$options .= "                      (default = 1000)\n";
			$options .= "         -i<f|n|p|s> = format of sequence file (fasta (f), nexus (n),\n";
			$options .= "                       phylip (p), or Se-Al (s)); default = autodetect\n";
			$options .= "         -k<number> = minimum k / theta ratio to define a species\n";
			$options .= "                      (default = 4)\n";
			$options .= "         -K = force K2P + G model\n";
			$options .= "         -N<number> = number of processors available for parallel processing\n";
			$options .= "                      (default = 1 (off))\n";
			$options .= "         -o<outgroup(s)> = name(s) of outgroup(s) for optimality analyses,\n";
			$options .= "                           with multiple names separated by a comma\n";
			$options .= "         -p<name> = name (including path) of PhyML 3.x program in \$PATH\n";
			$options .= "                    (default = $phyml)\n";
			$options .= "         -P<name> = name (including path) of PAUP* program in \$PATH\n";
			$options .= "                    (default = $paup)\n";
			$options .= "         -r<name> = name (including path) of RAxML program in \$PATH\n";
			$options .= "                    (default = $raxml)\n";
			$options .= "         -s<ML|NJ> = derive final tree using either NJ (with PAUP*) or ML\n";
			$options .= "                     (with RAxML; default)\n";
			$options .= "         -t<filename> = file containing optimal tree (with branch lengths)\n";
			$options .= "                        for sequence data in -d\n";
			$options .= "         -z = allow negative branch lengths for NJ analyses\n";
			$options .= "         -v = verbose output\n";
			$options .= "         -h = print this message and quit\n";

# Process switches
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] =~ /^-d(.*)/)
			{
			push @errorList, "ERROR: Sequence file $1 does not exist.\n" unless (-e $1);
			$dataFile = $1;
			}
		elsif ($ARGV[$i] =~ /^-a(.+)/)
			{
			$autoMT = $1;
			}
		elsif ($ARGV[$i] =~ /^-b(\d+)/)
			{
			$numBoot = $1;
			}
		elsif ($ARGV[$i] eq "-if")
			{
			$inputType = "fasta";
			}
		elsif ($ARGV[$i] eq "-in")
			{
			$inputType = "nexus";
			}
		elsif ($ARGV[$i] eq "-ip")
			{
			$inputType = "phylip";
			}
		elsif ($ARGV[$i] eq "-is")
			{
			$inputType = "Se-Al";
			}
		elsif ($ARGV[$i] =~ /^-k(\d+\.?\d*)/)
			{
			$kTheta = $1;
			push @errorList, "ERROR: k / theta ratio must be greater than 0\n" unless ($kTheta > 0);
			}
		elsif ($ARGV[$i] eq "-K")
			{
			$forceK2P = 1;
			}
		elsif ($ARGV[$i] =~ /-N([\d]+)/)
			{
			$numParallel = $1;
			$numParallel = 1 if ($numParallel < 1);
			}
		elsif ($ARGV[$i] =~ /^-o(.+)/)
			{
			$outGroups = "$1";
			}
		elsif ($ARGV[$i] =~ /^-p(.+)/)
			{
			$phyml = $1;
			}
		elsif ($ARGV[$i] =~ /^-P(.+)/)
			{
			$paup = $1;
			}
		elsif ($ARGV[$i] =~ /^-r(.+)/)
			{
			$raxml = $1;
			}
		elsif ($ARGV[$i] eq "-sML")
			{
			$searchStrategy = "ML";
			}
		elsif ($ARGV[$i] eq "-sNJ")
			{
			$searchStrategy = "NJ";
			}
		elsif ($ARGV[$i] =~ /^-t(.*)/)
			{
			push @errorList, "ERROR: Treefile $1 does not exist.\n" unless (-e $1);
			$treeFile = $1;
			}
		elsif ($ARGV[$i] eq "-v")
			{
			$verbose = 1;
			}
		elsif ($ARGV[$i] eq "-z")
			{
			$setZero = "setzero";
			}
		elsif ($ARGV[$i] eq "-x")
			{
			$debug = $verbose = 1;
			}
		elsif ($ARGV[$i] eq "-h")
			{
			print "$usage\n";
			print "Version: $version\n";
			print "$options\n";
			exit(0);
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "$usage\n";
			print "Version: $version\n";
			exit(1); 
			}
		}

	# Error checking
		push @errorList, "ERROR: Must specify file containing sequence data.\n" if (not defined $dataFile);

		# Check for PAUP*
			system ("$paup -h > paupTest.txt 2>&1");
				push @errorList, "ERROR: Could not call $paup.\n" if (-z "paupText.txt");
				unlink "paupTest.txt";

		# Check for RAxML
			if ($searchStrategy eq "ML")
				{
				system ("$raxml -h > raxmlTest.txt");
					if (-z "raxmlTest.txt")
						{
						push @errorList, "ERROR: $raxml could not be called successfully.\n";
						undef $raxml;
						}
					unlink "raxmlTest.txt";

				# Check for PTHREADS version
					if (defined $numParallel and defined $raxml)
						{
						system ("$raxml -T $numParallel > raxmlTest.txt");
						system ("grep Option raxmlTest.txt > threadTest.txt");
						$numThreads = (-s "threadTest.txt") ? 0 : $numParallel;
						unlink <raxmlTest.txt threadTest.txt>;
						push @errorList, "\tERROR: Have specified <2 threads ($numParallel) for PTHREADS version of RAxML.\n" if ($numThreads and $numParallel < 2);
						}
				}

		if (@errorList)
			{
			print "$_" foreach (@errorList);
			exit(1);
			}

		print "Script was called as (" . localtime() . "):\n\t$callString\n" if (defined $callString);

# Read in data matrix
	my ($dataSource, $dir, $ext) = fileparse($dataFile, qr/\.\D.*/);

	my $dataZero = time;
	seqRead($dataFile);
		$ntax = scalar(@accNum);
		foreach my $accession (@accNum)
			{
			$nameLabel{$accession} =~ s/\W+/_/g;
				$nameLabel{$accession} =~ s/\_+/_/g;
			$finalSeq{$accession} = $sequence{$accession};
			$maxLength = length($finalSeq{$accession}) if (length($finalSeq{$accession}) > $maxLength);
			$namePresent{$nameLabel{$accession}} = 1;
			}

	# Print working files to disk
		$nexusPrint = 1;
			$nexusOut = "userData_fourX.nex";
		$phylipExtPrint = 1;
			$phylipOut = "userData_fourX.phylip";

		seqPrint($dataFile);

	# Check outgroups
		my $newOutGroups;
		if (defined $outGroups)
			{
			my @OGtaxa;
				foreach my $OG (split(/,/, $outGroups))
					{
					push @OGtaxa, $OG if (defined $namePresent{$OG});
					}
					
				$newOutGroups = join(',', @OGtaxa);
					$newOutGroups = "<empty set>" unless ($newOutGroups);
			
			print "\t\tWARNING: outgroups redfined to $newOutGroups\n" if (defined $newOutGroups and defined $outGroups and $newOutGroups ne $outGroups);
			}

	printf "\tTime taken: %s seconds\n", time - $dataZero;

# Read in / get phylogenetic tree
	my @allTrees;
	my $userTreeFile = "userTree_fourX.tre";
	my $bootNodeTree;
	
	if (defined $treeFile and -e $treeFile)
		{
		print "\nReading in user tree from file $treeFile ...\n";
			my $treeZero = time;
		
		@allTrees = treeRead($treeFile);
			die "ERROR: No trees read in from file $treeFile\n" unless (@allTrees);
			
		open (TREE, ">$userTreeFile") or die "Cannot write user tree to nexus file $userTreeFile ($!)\n";
			print TREE "#NEXUS\n\n";
			print TREE "begin trees;\n";
			print TREE "\t tree $treeName{$allTrees[0]} = $rootingStatus{$allTrees[0]} $treeDescription{$allTrees[0]};\n";
			print TREE "end;\n";
		close TREE;
	
		printf "\tTime taken: %s seconds\n", time - $treeZero;
		}
	elsif ($searchStrategy eq "ML")
		{
		print "WARNING: Tree file $treeFile does not exist.\n" if (defined $treeFile);
		print "\nDetermining ML tree using $raxml ...\n";
			my $raxmlZero = time;
			
		my $runID = "$dataSource\_boot";
			unlink <RAxML_*$runID>;
		my $logName = "fourX_raxmlLog.txt";

		my $bootSeed = int(rand(10000)+1);
		my $raxmlBootcmd = "$raxml -s $phylipOut -n $runID -m GTRCAT -f a -N $numBoot -x $bootSeed";
			$raxmlBootcmd .= " -o $newOutGroups" if (defined $newOutGroups and $newOutGroups ne "<empty set>");
#			my $newOutGroups;
#			if (defined $outGroups)
#				{
#				my @OGtaxa;
#					foreach my $OG (split(/,/, $outGroups))
#						{
#						push @OGtaxa, $OG if (defined $namePresent{$OG});
#						}
#						
#					$newOutGroups = join(',', @OGtaxa);
#						$newOutGroups = "<empty set>" unless ($newOutGroups);
#
#				$raxmlBootcmd .= " -o $newOutGroups" unless ($newOutGroups eq "<empty set>");
#				}

			$raxmlBootcmd .= " -T $numThreads" if ($numThreads > 1);
			$raxmlBootcmd .= " > $logName";
			print "\tUsing call: $raxmlBootcmd\n";
#				print "\t\tWARNING: outgroups redfined to $newOutGroups\n" if (defined $newOutGroups and defined $outGroups and $newOutGroups ne $outGroups);

		# Do RAxML analysis
			system ("$raxmlBootcmd");
				printf "\t\tTime taken: %s seconds\n", time - $raxmlZero;
				die "ERROR: RAxML analyses do not appear to have been completed.\n" unless (-e "RAxML_bestTree.$runID");

		# Associate bootstrap values to optimal tree
			my $bootFile = "RAxML_bootstrap.$runID";
			my $optTreeFile = "RAxML_bestTree.$runID";
				(my $stripFile = $optTreeFile) =~ s/\.RUN\.\d+$//;
				$stripFile .= "_bootstrip.tre";
			my ($bootLengthTree, $bootNodeTree);
		
			$keepLengths = 0;
			BootStrip("$runID -b$bootFile -d1 -fl -t$optTreeFile");
				setLineBreak($stripFile);
				open (STRIP, "<$stripFile") or die "Cannot open result from BootStrip analysis, $stripFile ($!).\n";
					while (<STRIP>)
						{
						chomp;
						$bootLengthTree = $1 if ($_ =~ /^\s+tree\sreadTree_\d+\s+\=\s+\[&U\]\s+(.+)/i);
						}
				close STRIP;
				$bootLengthTree =~ s/;$//;

			BootStrip("$runID -b$bootFile -d1 -fn -k -t$optTreeFile");
				setLineBreak($stripFile);
				open (STRIP, "<$stripFile") or die "Cannot open result from BootStrip analysis, $stripFile ($!).\n";
					while (<STRIP>)
						{
						chomp;
						$bootNodeTree = $1 if ($_ =~ /^\s+tree\sreadTree_\d+\s+\=\s+\[&U\]\s+(.+)/i);
						}
				close STRIP;
				$bootNodeTree =~ s/;$//;
				$bootNodeTree =~ s/\'n\/a\'//g;
				$bootNodeTree =~ s/\'</\'/g;
			unlink $stripFile;
			
		# Print trees to file
			@allTrees = treeRead($optTreeFile);
				die "ERROR: No trees read in from file $optTreeFile\n" unless (@allTrees);
				$treeName{$allTrees[0]} = "$dataSource";
				$rootingStatus{$allTrees[0]} = "[&R]" if (defined $outGroups);

			open (TREE, ">$userTreeFile") or die "Cannot write user tree to nexus file $userTreeFile ($!)\n";
				print TREE "#NEXUS\n\n";
				print TREE "begin trees;\n";
				print TREE "\t tree $treeName{$allTrees[0]} = $rootingStatus{$allTrees[0]} $treeDescription{$allTrees[0]};\n";
				print TREE "\t tree $dataSource\_bootNode = $rootingStatus{$allTrees[0]} $bootNodeTree;\n";
				print TREE "\t tree $dataSource\_bootLength = $rootingStatus{$allTrees[0]} $bootLengthTree;\n";
				print TREE "end;\n";
			close TREE;
			
		# Clean up RAxML analyses
			mkdir "RAxML" unless (-d "RAxML");
			
			my $mvCommand = "mv RAxML_* RAxML";
				$mvCommand =~ s/^mv -f/^move/ if ($Config{osname} =~ /^MSWin/i or $OSNAME =~ /^MSWin/i);
				system ("$mvCommand");

			my $copyCommand = "cp -P $userTreeFile RAxML/$dataSource\_RAxML.tre";
				$copyCommand =~ s/^cp -P/^copy/ if ($Config{osname} =~ /^MSWin/i or $OSNAME =~ /^MSWin/i);
				system ("$copyCommand");
				
			unlink $logName;
		}

# Do ModelTEST using user-input / ML tree or from scratch
	unless (-e "autoMTresults.txt")
		{
		my $modelScheme = "-nd";
		if ($forceK2P)
			{
			print "\nDetermining parameters for K2P + G model using $autoMT ...\n";
			$modelScheme = "-nk";
			}
		else
			{
			print "\nDetermining optimal model using $autoMT ...\n";
			}
		
		my $autoCommand = "$autoMT -f$nexusOut -in $modelScheme -N$numParallel -p$phyml -P$paup";
			$autoCommand .= " -t$userTreeFile" if (-e $userTreeFile and (defined $treeFile or $searchStrategy eq "ML"));

		system ($autoCommand);
			die "ERROR: $autoMT does not appear to have launched properly\n" unless (-e "autoMTresults.txt");
		}
		
	# Read in and parse autoMT output
		my $optModel;
		my $lsetCommand;
	
		setLineBreak("autoMTresults.txt");
		open (MT, "<autoMTresults.txt") or die "Cannot open results file from autoMT.pl analyses ($!)\n";
			while (<MT>)
				{
				chomp;
				next unless ($_);
				next unless ($_ =~ /^$nexusOut/);

				my @autoLine = split(/\t/, $_);
					$optModel = $autoLine[1];
						$optModel =~ s/\s+\*$//;
					$lsetCommand = pop @autoLine;
				}
		close MT;

		# Parse model
			my $plusI;
				$plusI = $1 if ($optModel =~ /\+I$/ and $lsetCommand =~ /pinvar\s*=\s*(\d\.\d+)/);
			
			my $plusG = "equal";
				$plusG = "gamma shape=$1" if ($optModel =~ /\+G$/ and $lsetCommand =~ /shape\s*=\s*(\d+\.\d+)/);

			$optModel =~ s/\+I//;
			$optModel =~ s/\+G//;
			$optModel = "HKY85" if ($optModel eq "HKY");
			$optModel = "TamNei" if ($optModel eq "TrN");

# Do NJ analysis (if applicable)
	if ($searchStrategy eq "NJ")
		{
		print "\nDetermining NJ tree using $paup ...\n";
			my $njZero = time;

		my $commandFile = "NJcommandFile.txt";
		open (CMD, ">$commandFile") or die "Cannot write to PAUP* command file $commandFile ($!)\n";
			print CMD "\nbegin paup;\n";
			print CMD "\tset taxlabels = full criterion = distance;\n";
			if (defined $newOutGroups and $newOutGroups ne "<empty set>")
				{
				$newOutGroups =~ s/,/ /g;
				print CMD "\toutgroup $newOutGroups;\n";
				}
			print CMD "\n\tdset negbrlen = $setZero distance=$optModel";
				print CMD " pinvar = $plusI" if (defined $plusI);
				print CMD " rates = $plusG;\n";
			print CMD "\tnj;\n";
			print CMD "\t\tsavetrees format = altnexus file = $dataSource\_njTree.tre brlens = yes replace;\n";
			print CMD "\n\tlog file = $dataSource\_njBootlog.txt replace;\n";
			print CMD "\t\tbootstrap nreps = $numBoot search = nj keepall = yes;\n";
			print CMD "\tlog stop;\n";
			print CMD "end;\n";
			print CMD "\nquit warntsave=no;\n";
		close CMD;
	
		system ("cat $nexusOut $commandFile > NJmatrix.nex");
		system ("$paup NJmatrix.nex > NJtree_log.txt");

		# Associate bootstrap values to optimal tree
			my $bootLog = "$dataSource\_njBootlog.txt";
			my $optTreeFile = "$dataSource\_njTree.tre";
				my $stripFile = $optTreeFile . "_bootstrip.tre";
			my ($bootLengthTree, $bootNodeTree);
		
			$keepLengths = 0;
			BootStrip("NJ -l$bootLog -d1 -fl -t$optTreeFile");
				setLineBreak($stripFile);
				open (STRIP, "<$stripFile") or die "Cannot open result from BootStrip analysis, $stripFile ($!).\n";
					while (<STRIP>)
						{
						chomp;
						$bootLengthTree = $1 if ($_ =~ /^\s+tree\sPAUP_\d+\s+\=\s+\[&U\]\s+(.+)/i);
						}
				close STRIP;
				$bootLengthTree =~ s/;$//;

			BootStrip("NJ -l$bootLog -d1 -fn -k -t$optTreeFile");
				setLineBreak($stripFile);
				open (STRIP, "<$stripFile") or die "Cannot open result from BootStrip analysis, $stripFile ($!).\n";
					while (<STRIP>)
						{
						chomp;
						$bootNodeTree = $1 if ($_ =~ /^\s+tree\sPAUP_\d+\s+\=\s+\[&U\]\s+(.+)/i);
						}
				close STRIP;
				$bootNodeTree =~ s/;$//;
				$bootNodeTree =~ s/\'n\/a\'//g;
				$bootNodeTree =~ s/\'</\'/g;
			unlink $stripFile;
			
		# Print trees to file
			@allTrees = treeRead($optTreeFile);
				die "ERROR: No trees read in from file $optTreeFile\n" unless (@allTrees);
				$treeName{$allTrees[0]} = "$dataSource";
				$rootingStatus{$allTrees[0]} = "[&R]" if (defined $outGroups);

			open (TREE, ">$userTreeFile") or die "Cannot write user tree to nexus file $userTreeFile ($!)\n";
				print TREE "#NEXUS\n\n";
				print TREE "begin trees;\n";
				print TREE "\t tree $treeName{$allTrees[0]} = $rootingStatus{$allTrees[0]} $treeDescription{$allTrees[0]};\n";
				print TREE "\t tree $dataSource\_bootNode = $rootingStatus{$allTrees[0]} $bootNodeTree;\n";
				print TREE "\t tree $dataSource\_bootLength = $rootingStatus{$allTrees[0]} $bootLengthTree;\n";
				print TREE "end;\n";
			close TREE;
			
		# Clean up RAxML analyses
			mkdir "NJ" unless (-d "NJ");
			
			my $mvCommand = "mv $commandFile NJtree_log.txt NJ";
				$mvCommand =~ s/^mv -f/^move/ if ($Config{osname} =~ /^MSWin/i or $OSNAME =~ /^MSWin/i);
				system ("$mvCommand");

			my $copyCommand = "cp -P $userTreeFile NJ/$dataSource\_NJ.tre";
				$copyCommand =~ s/^cp -P/^copy/ if ($Config{osname} =~ /^MSWin/i or $OSNAME =~ /^MSWin/i);
				system ("$copyCommand");
				
			unlink "$dataSource\_njLog.txt NJmatrix.nex $dataSource\_njTree.tre $dataSource\_njBootlog.txt";
			
		printf "\tTime taken: %s seconds\n", time - $njZero;
		}

# Derive all pairwise distances
	# Derive uncorrected distances
		my $commandFile = "pCommandFile.txt";	
		open (CMD, ">$commandFile") or die "Cannot write to PAUP* command file $commandFile ($!)\n";
			print CMD "\nbegin paup;\n";
			print CMD "\tset taxlabels = full;\n";
			print CMD "\n\ttstatus full;\n";
			print CMD "\n\tdset distance=p;\n";
			print CMD "\tshowdist;\n";
			print CMD "end;\n";
			print CMD "\nquit warntsave=no;\n";
		close CMD;

		system ("cat $nexusOut $commandFile > pMatrix.nex");
		system ("$paup pMatrix.nex > pDistances_log.txt");
		matrixRead("pDistances_log.txt");
			my %pDistances = %belowDiag;
			undef %belowDiag;

	# Derive corrected distances
		$commandFile = "kCommandFile.txt";	
		open (CMD, ">$commandFile") or die "Cannot write to PAUP* command file $commandFile ($!)\n";
			print CMD "\nbegin paup;\n";
			print CMD "\tset taxlabels = full;\n";
			print CMD "\n\ttstatus full;\n";
			print CMD "\n\tdset distance=$optModel";
				print CMD " pinvar = $plusI" if (defined $plusI);
				print CMD " rates = $plusG;\n";
			print CMD "\tshowdist;\n";
			print CMD "end;\n";
			print CMD "\nquit warntsave=no;\n";
		close CMD;
	
		system ("cat $nexusOut $commandFile > kMatrix.nex");
		system ("$paup kMatrix.nex > kDistances_log.txt");
	
		matrixRead("kDistances_log.txt");
			my %kDistances = %belowDiag;
			undef %belowDiag;

	if ($debug)
		{
		for (my $i = 0; $i < scalar(@allTaxa) - 1; $i++)
			{
			for (my $j = $i + 1; $j < scalar(@allTaxa); $j++)
				{
				print "$allTaxa[$i]\t$allTaxa[$j]\t$pDistances{$allTaxa[$i]}{$allTaxa[$j]}\t$kDistances{$allTaxa[$i]}{$allTaxa[$j]}\n";
				}
			}
		}

# Determine 4x status over all clades in user-input tree
	print "\nDetermining 4x clades on tree $treeName{$allTrees[0]} ... \n";
		my $fourZero = time;

	parseTree($allTrees[0]);

	my (%sumP, %sumK);
	
	my $resultsFile = "$dataSource\_fourXresults.txt";
		open (OUT, ">$resultsFile") or die "Cannot write final tree to $resultsFile ($!)";
		print OUT "cladeID\taverage P (clade)\tminimum K\tComparison\tmaximum P (daughter)\tDaughter\tTheta\tK / Theta\tSignificant\n";
	
	foreach my $cladeID (reverse @nodeList)
		{
		unless ($cladeID =~ /^GTnode/)	# Set sum of p-distances for terminal taxa to zero
			{
			$sumP{$cladeID} = $aveP{$cladeID} = 0;
			push @{ $geneNodeMember{$cladeID} }, $cladeID;
			next;
			}
			print "\n\t$cladeID\n" if ($verbose);

		# Get sum of p-distances for clade
			print "\t\tCalculating p-distances " if ($debug);
			
			my $numClades = scalar(@nodeList) - scalar(@GTtaxa);

			if ($numClades != scalar(@GTtaxa) - 1)	# Tree is not binary; determine using brute force approach that determines de novo for each clade; less efficient but allows for polytomies
				{
				print "(via brute force method) ...\n" if ($debug);
				for (my $i = 0; $i < scalar(@{ $geneNodeMember{$cladeID} }) - 1; $i++)
					{
					print "\t\t\tComparing $geneNodeMember{$cladeID}[$i] vs:\n" if ($debug);
					for (my $j = $i + 1; $j < scalar(@{ $geneNodeMember{$cladeID} }); $j++)
						{
						print "\t\t\t\t$geneNodeMember{$cladeID}[$j] ($pDistances{$geneNodeMember{$cladeID}[$i]}{$geneNodeMember{$cladeID}[$j]})\n" if ($debug);
						$sumP{$cladeID} += $pDistances{$geneNodeMember{$cladeID}[$i]}{$geneNodeMember{$cladeID}[$j]};
						}
					}
				}
			else	# Elegant (faster?) approach that re-uses nested values; but assumes tree is binary
				{
				print "(via elegant method) ...\n" if ($debug);
				$sumP{$cladeID} += $sumP{$_} foreach (@ {$daughterClade{$cladeID} });
	
				# Check if sum of p-distances defined for all daughter taxa
					my $sumCheck = 0;
					foreach my $daughterID (@ {$daughterClade{$cladeID} })
						{
						print "\t\t\tdaughterID: $daughterID ($cladeSize{$daughterID})\n" if ($debug);
						$sumCheck++ if (defined $sumP{$daughterID});
						}
					print "\t\t\t$sumCheck\n" if ($debug);
					
				if ($sumCheck == scalar(@ {$daughterClade{$cladeID} }))	#	Sum of p-distances defined for both daughter clades; only need to compare species between clades
					{
					foreach my $i (@{ $geneNodeMember{$daughterClade{$cladeID}[0]} })
						{
						print "\t\t\tComparing $i vs:\n" if ($debug);
						foreach my $j (@{ $geneNodeMember{$daughterClade{$cladeID}[1]} })
							{
							print "\t\t\t\t$j ($pDistances{$i}{$j})\n" if ($debug);
							$sumP{$cladeID} += $pDistances{$i}{$j};
							}
						}
					}
				else	# Need to do all pairwise comparisons within clade
					{
					for (my $i = 0; $i < scalar(@{ $geneNodeMember{$cladeID} }) - 1; $i++)
						{
						print "\t\t\tComparing $geneNodeMember{$cladeID}[$i] vs:\n" if ($debug);
						for (my $j = $i + 1; $j < scalar(@{ $geneNodeMember{$cladeID} }); $j++)
							{
							print "\t\t\t\t$geneNodeMember{$cladeID}[$j] ($pDistances{$geneNodeMember{$cladeID}[$i]}{$geneNodeMember{$cladeID}[$j]})\n" if ($debug);
							$sumP{$cladeID} += $pDistances{$geneNodeMember{$cladeID}[$i]}{$geneNodeMember{$cladeID}[$j]};
							}
						}
					}
				}
			print "\t\t\tsumP = $sumP{$cladeID}\n" if ($debug);

			my $numP = ($cladeSize{$cladeID}**2 - $cladeSize{$cladeID}) / 2;
				$aveP{$cladeID} = $sumP{$cladeID} / $numP;
				if ($verbose)
					{
					printf "\t\tAverage pairwise P-distance within clade: %.3f (%.3f / $numP)\n", $aveP{$cladeID}, $sumP{$cladeID};
					}

		# Get k-distances between daughter clades
			print "\n\t\tCalculating k-distances ... \n" if ($debug);
				my ($minK, @minKDaughters, $minKDaughters);
				foreach my $i (@{ $geneNodeMember{$daughterClade{$cladeID}[0]} })
					{
					print "\t\t\tComparing $i vs:\n" if ($debug);
					foreach my $j (@{ $geneNodeMember{$daughterClade{$cladeID}[1]} })
						{
						print "\t\t\t\t$j ($kDistances{$i}{$j})\n" if ($debug);
						$sumK{$cladeID} += $kDistances{$i}{$j};
						if (not defined $minK or $kDistances{$i}{$j} <= $minK)
							{
							$minK = $kDistances{$i}{$j};
							push @minKDaughters, "$i,$j";
							}
						}
					}
				$minKDaughters = join(';', @minKDaughters);
				
				print "\t\t\tsumK = $sumK{$cladeID}\n" if ($debug);
				print "\t\t\tminK = $minK ($minKDaughters)\n" if ($debug);

				my $numK = $cladeSize{$daughterClade{$cladeID}[0]} * $cladeSize{$daughterClade{$cladeID}[1]};
					my $aveK = $sumK{$cladeID} / $numK;
					if ($verbose)
						{
#						printf "\t\tAverage pairwise K-distance between daughter clades: %.3f (%.3f / $numK)\n", $aveK, $sumK{$cladeID};
						printf "\t\tMinimum pairwise K-distance between daughter clades: %.3f ($minKDaughters)\n", $minK;
						}

		# Calculate K / theta ratio
			print "\t\tCalculating K / theta ratio\n" if ($debug);
				
			# Find daughter clade with maximum within clade P-distance
				my $maxP = -1;
				my @maxDaughter;
				foreach my $daughterID (@ {$daughterClade{$cladeID} })
					{
					 if (defined $aveP{$daughterID} and $aveP{$daughterID} >= $maxP)
						{
						$maxP = $aveP{$daughterID};
						print "\t\t\tMaximum P-distance = $maxP ($daughterID)\n" if ($debug);
						undef @maxDaughter if ($aveP{$daughterID} >= $maxP);
							push @maxDaughter, $daughterID;
						}
					$kThetaStatus{$cladeID} = "nested" if (defined $kThetaStatus{$daughterID});
					}

			(my $printID = $cladeID) =~ s/^GTnode//;
			my $printDaughter = join(',', @maxDaughter);
				$printDaughter =~ s/GTnode//g if (@maxDaughter);
					
			if ($maxP == 0)
				{
				$kTheta{$cladeID} = "undefined";
				print "\t\tK / theta ratio = $kTheta{$cladeID}\n" if ($verbose);
#				printf OUT "$printID\t%.5f\t%.5f\t$maxP ($printDaughter)\t0\t$kTheta{$cladeID}\n", $aveP{$cladeID}, $aveK;
				printf OUT "$printID\t%.5f\t%.5f\t$minKDaughters\t$maxP\t$printDaughter\t0\t$kTheta{$cladeID}\n", $aveP{$cladeID}, $minK;
				}
			else
				{
#				$kTheta{$cladeID} = $aveK / ($maxP / (1 - (4 * $maxP / 3)));
				$kTheta{$cladeID} = $minK / ($maxP / (1 - (4 * $maxP / 3)));
					printf "\n\t\t\ttheta  = %s\n", $maxP / (1 - (4 * $maxP / 3)) if ($debug);

#				printf OUT "$printID\t%.5f\t%.5f\t%.5f ($printDaughter)\t%.5f\t%.5f", $aveP{$cladeID}, $aveK, $maxP, $maxP / (1 - (4 * $maxP / 3)), $kTheta{$cladeID};
				printf OUT "$printID\t%.5f\t%.5f\t$minKDaughters\t%.5f\t$printDaughter\t%.5f\t%.5f", $aveP{$cladeID}, $minK, $maxP, $maxP / (1 - (4 * $maxP / 3)), $kTheta{$cladeID};
					if ($kTheta{$cladeID} >= $kTheta)
						{
						$kThetaStatus{$cladeID} = "primary" if (not defined $kThetaStatus{$cladeID});
						print OUT (defined $kThetaStatus{$cladeID} and $kThetaStatus{$cladeID} eq "primary") ? "\t*" : "\t(*)";
						}
					print OUT "\n";

				if ($verbose)
					{
					printf "\t\tK / theta ratio = %.5f", $kTheta{$cladeID};
						print STDOUT (defined $kThetaStatus{$cladeID} and $kThetaStatus{$cladeID} eq "primary") ? " *" : " (*)";
						print "\n";
					}
				}
		}
	close OUT;
		
	# Reparse tree and print to file
		my $treeGrow;
		my $geneTree = (defined $bootNodeTree) ? $bootNodeTree : $treeDescription{$allTrees[0]};
		my ($position, $cladeNum, $CurrentClade) = (0, 0, 0);
		my $OldCurrent;
		my $taxon;
		my $cladeID;

		while ($position < length($geneTree))
			{
			my $offset = 1;
			if (substr($geneTree, $position, 1) eq "(")		# New clade; go one level deeper
				{
				$cladeNum++;
					$cladeID = "GTnode" . $cladeNum;
				$CurrentClade = $cladeID;
				$treeGrow .= "(";
				}
			elsif (substr($geneTree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
				{
				$OldCurrent = $CurrentClade;
				$CurrentClade = $ParentClade{$OldCurrent};
				# Determine new branch length
#					my $branchLength = sprintf "%.5f", $cladeDepth{$gene2ST{$OldCurrent}} -  $cladeDepth{$gene2ST{$CurrentClade}};
				$treeGrow  .= ")";
					if ($CurrentClade ne "GTnode0" and $kTheta{$CurrentClade} ne "undefined" and $kTheta{$CurrentClade} >= $kTheta)
						{
						$treeGrow .= ($kThetaStatus{$CurrentClade} eq "primary") ? "[&!color=#-16711885]" : "[&!color=#-65536]";
						}

				(my $cladeNum = $OldCurrent) =~ s/^GTnode//;
					$treeGrow .= "'$cladeNum'";
				$treeGrow .= ":$CladeSupport{$OldCurrent}" if (defined $CladeSupport{$OldCurrent});
				}
			elsif (substr($geneTree, $position, 1) eq ":")	# Store branch lengths for node or taxon
				{
				until (($position+$offset) > length($geneTree) or substr($geneTree, $position+$offset, 1) eq "(" or substr($geneTree, $position+$offset, 1) eq ")" or substr($geneTree, $position+$offset, 1) eq ",")
					{
					$offset++;
					}
				$position += $offset-1;
				}
			elsif (substr($geneTree, $position, 1) eq ",")
				{
				$treeGrow .= ",";
				}
			else	# Determine membership of clade
				{
				until (substr($geneTree, $position+$offset, 1) eq "(" or substr($geneTree, $position+$offset, 1) eq ")" or substr($geneTree, $position+$offset, 1) eq "," or substr($geneTree, $position+$offset, 1) eq ":")
					{
					$offset++;
					}
				$taxon = substr($geneTree, $position, $offset);
#				print "\t\t\tTaxon: $taxon\n" if ($debug);
#				my $branchLength = sprintf "%.5f", $cladeDepth{$taxon} -  $cladeDepth{$gene2ST{$CurrentClade}};
					$treeGrow .= "$taxon";
						$treeGrow .= "[&!color=#-16711885]" if ($kTheta{$CurrentClade} ne "undefined" and $kTheta{$CurrentClade} >= $kTheta);
					$treeGrow .= ":$CladeSupport{$taxon}" if (defined $CladeSupport{$taxon});
				$position += $offset-1;
				}
			$position++;
			}

		$treeGrow =~ s/,+/,/g;
		$treeGrow =~ s/,\)/\)/g;
		$treeGrow =~ s/\'n\/a\'//g,
	
		my $fourXfile = "$dataSource\_fourX.tre";
		open (FOURX, ">$fourXfile") or die "Cannot write final tree to $fourXfile ($!)";
			print FOURX "#NEXUS\n\n";
			print FOURX "begin trees;\n";
			print FOURX "\t tree $treeName{$allTrees[0]}\_fourX = $rootingStatus{$allTrees[0]} $treeGrow;\n";
			print FOURX "end;\n";
		close FOURX;

	printf "\n\tTime taken %s seconds\n", time - $fourZero;
	
	# Clean up
		unless ($debug)
			{
			unlink <kCommand* kDistances* kMatrix*>;
			unlink <pCommand* pDistances* pMatrix*>;
			unlink <userData_fourX* userTree_fourX*>;
			unlink "autoMT_optTrees.tre";
			}

printf "\nTotal time taken: %s seconds\n", time - $dataZero;

exit(0);

# Subroutines used in script

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

sub duplStrip	# Subroutine to strip duplicate entries from an array
	{
	my @duplArray = @_;
	my %dupl;		# Remove duplicate taxa
		my @stripArray = grep(!$dupl{$_}++, @duplArray);
		@stripArray = sort {$a cmp $b} @stripArray;
	return @stripArray;
	}

sub treeRead	# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
	{
	my $treeFile = shift;
	my $tntFlag = 0;

	setLineBreak($treeFile);
	open (TREES, "<$treeFile") or die "Cannot open nexus-formatted treefile $treeFile ($!)\n";
		my (@allTrees, $rawTreeDescrip, $treeDescrip, $treeFlag, $treeCount, $rooting);
		my (@taxonList, %taxonPresent);
		my (%TTable, $translation, $translateFlag,$maxTaxonID, $transText);
		my $commentFlag = 0;
		
		while (<TREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			$_ =~ s/\s+$//;		# Strip (any) trailing white space
			
			# Block any commented lines
				if ($_ =~ /^\[/)
					{
					$commentFlag += ($_ =~ tr/\[+//);
					}
				if ($_ =~ /\]\s?$/)
					{
					$_ =~ s/\s+$//;
					$commentFlag -= ($_ =~ tr/\]+//);
					next;
					}
				next if ($commentFlag > 0);
				
			# Process various TNT commands
				if ($_ =~ /^tread/i)		# Clear variables for each tree block
					{
					undef %TTable;
					$translation = $translateFlag = $treeFlag = $treeCount = 0;
					$tntFlag = 1;
					$treeBlock++;
					}
				last if ($_ =~ /^proc-/i);

			# Process various nexus commands
				if ($_ =~ /^begin trees;/i)		# Clear variables for each tree block
					{
					undef %TTable;
					$translation = $translateFlag = $treeFlag = $treeCount = 0;
					$treeBlock++;
					}
				if ($_ =~ /^translate/i)	# Read translation table of taxon names (if any)
					{
					print "Reading translation table ...\n" if ($verbose);
					$translateFlag = $translation = 1;
					$transText = "";
					$maxTaxonID = 0;
					}
				if ($translateFlag)
					{
					next if ($_ =~ /^translate/i);
					$translateFlag = 0 if ($_ =~ /;/);	# End of translate comment
					
					$transText = $_;
						$transText =~ s/;\s?$//;
						$transText =~ s/,\s?$//;
						$transText =~ s/^\s+//;	# Strip (any) leading white space
						next if ($transText eq "");
						
					my @transEntry = split(/\s/, $transText);
						my $taxonID = shift(@transEntry);
						my $taxonName = join('_', @transEntry);

					$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
					$taxonName =~ s/\s+/_/g;
						$taxonName =~ s/\W+/_/g;
						$taxonName =~ s/\_+/_/g;
						$taxonName =~ s/^_//;
						$taxonName =~ s/_$//;
						$taxonName = "'" . $taxonName . "'" if ($taxonName =~ /\W/);
						$taxonName =~ s/^\'+/\'/;
						$taxonName =~ s/\'+$/\'/;
					$TTable{$taxonID} = $taxonName;
					$taxonPresent{$taxonName}++;
					push @taxonList, $taxonName;
					}
				if (($_ =~ /tree/i and $_ =~ /\=/) or $_ =~ /^\(/)
					{
					unless ($_ =~ /^>/)	# Ignore (any) commented lines
						{
						$rawTreeDescrip = $treeDescrip = "" unless ($treeFlag);	# Allows tree to be spread over multiple lines
						unless ($block == 0 and $treeBlock > 1)
							{
							$treeFlag = 1;
							}
						}
					}
	
				if ($treeFlag)
					{
					$rawTreeDescrip .= $_;
						print "$rawTreeDescrip\n" if ($debug);
					if ($_ =~ /:/)
						{
						$supportFlag = 1;
						}
					if ($_ =~ /;/ or ($tntFlag and $_ =~ /\*/))	# End of tree description; parse
						{
						$treeCount++;
						print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
						print "\tTranslating taxon names ...\n" if ($verbose and $translation);
						# Determine if trees are rooted or not (default for no information is not rooted)
							if ($rawTreeDescrip =~ /\[&R\]/)
								{
								$rooting = "[&R]";
								}
							elsif ($rawTreeDescrip =~ /\[&U\]/)
								{
								$rooting = "[&U]";
								$unrootedFlag = 1;
								}
							else
								{
								$rooting = $defaultRooting;
								$unrootedFlag = 1 if ($defaultRooting eq "[&U]");
								}

						# Remove any embedded comments
							if ($rawTreeDescrip =~ /\[/ and not $keepComments)
								{
								for (my $position = 0; $position < length($rawTreeDescrip); $position++)
									{
									if (substr($rawTreeDescrip, $position, 1) eq "[")
										{
										$position++ until (substr($rawTreeDescrip, $position, 1) eq "]");
										}
									else
										{
										$treeDescrip .= substr($rawTreeDescrip, $position, 1);
										}
									}
								}
							else
								{
								$treeDescrip = $rawTreeDescrip;
								}
						$treeFlag = 0;
						
						# Pull out tree description
							my $position = 0;
							until (substr($treeDescrip, $position, 1) eq "(")		# Avoid anything before tree description proper
								{
								$position++;
								}
							my $tree = substr($treeDescrip, $position);
								# Fix TNT-formatted trees
									if ($tntFlag)
										{
										$tree =~ s/\s*\=/:/g;
										$tree =~ s/\s+/,/g;
										$tree =~ s/\)\(/\),\(/g;
										$tree =~ s/,\)/\)/g;										
										$tree =~ s/\*\s*$//;
										}
								$tree =~ s/;\s*$//;
								$tree =~ s/\s+/_/g;
								
						# Get name of tree (if any)
							my $name;
							if ($treeDescrip =~ /=/)
								{
								($name) = (split /=/, $treeDescrip, 0);
									$name =~ s/\s*\w*tree\s+//i;
									$name =~ s/\s*\*\s*//i;
									$name =~ s/\s+$//g;
									$name =~ s/-/_/g;
									$name =~ s/\s+/_/g;
									$name =~ s/^'//;
									$name =~ s/'$//;
									$name = "'" . $name . "'" if ($name =~ /\W/);
#									$treeNames{$treeCount} = $name;
								}
								
								
						# Parse tree description
							my $openCount = ($tree =~ tr/(//);
							my $closedCount = ($tree =~ tr/)//);
							my $taxonRead = 0;
							if ($openCount == $closedCount)		# Tree is balanced; process
								{
								$position = length($tree) - 1;
								while ($position >= 0)
									{
									my $IDlength = 1;
									if (not $taxonRead and (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":"))	# Tree structure; skip to next character
										{
										$position--;
										}
									else	# Read in taxonID
										{
										until (not $taxonRead and (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":"))
											{
											if (substr($tree, $position, 1) eq "'")
												{
												$taxonRead = ($taxonRead) ? 0 : 1;
												}
											$position--;
											$IDlength++;
											}
										unless (substr($tree, $position, 1) eq ":" or substr($tree, $position, 1) eq ")")		# Avoid reading branch lengths or node labels as taxon names
											{
											my $readName = substr($tree, $position+1, $IDlength-1);
											if ($translation and not defined $taxonPresent{$readName})	# Translate taxon names; latter condition is a safety for BioPerl convert_trees.pl module which adds, but doesn't use translation block
												{
												if ($TTable{$readName})
													{
													substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		# Translate name
													unless ($taxonPresent{$TTable{$readName}})
														{
														push @taxonList, $TTable{$readName};
														$taxonPresent{$TTable{$readName}} = 1;
														}
													}
												else
													{
													print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
													substr($tree, $position+1, $IDlength-1) = "xxx";		# Translate name
													unless ($taxonPresent{"xxx"})
														{
														push @taxonList, "xxx";
														$taxonPresent{"xxx"} = 1;
														}
													}
												}
											else	# Enter name as read from tree, unless already present
												{
												# Check for necessity of single quotes and correct as required
													if ($readName =~ /^'/ and $readName =~ /'$/)
														{
														my $newName = $readName;
															$newName =~ s/^'//;
															$newName =~ s/'$//;
															$newName = "'$newName'" if ($newName =~ /\W/);
														if ($newName ne $readName)
															{
															$tree =~ s/$readName,/$newName,/;
															$tree =~ s/$readName:/$newName:/;
															$tree =~ s/$readName\)/$newName\)/;
															$readName = $newName;
															}
														}
												$taxonPresent{$readName}++;
												push @taxonList, $readName if ($taxonPresent{$readName} == 1);
												}
											}
										}
									}
								
								$treeNum++;
									my $treeID = "treeID_$treeNum";
									$name = "readTree_$treeNum" if (not defined $name);
								
								push @allTrees, $treeID;
									$treeDescription{$treeID} = $tree;
									$rootingStatus{$treeID} = $rooting;
									$treeName{$treeID} = $name;
								print "\tTree description:\n\t\t$tree\n" if ($verbose);
								}
							else	# Tree is unbalanced; reject
								{
								print "$tree\n" if ($debug);
								print "\tTree #$treeCount of tree block #$treeBlock is unbalanced. Not processed.\n";
								}
						}
					}
			}
	close TREES;
	
	# Process taxon list
		@allTaxa = duplStrip(@taxonList);		
	
	return @allTrees;
	}

sub seqRead
	{
	my $seqFile = shift;
	undef %sequence;

	print "\nReading in sequence data from file $seqFile (type is $inputType) ...\n" if ($inputType and not $quiet);
	setLineBreak($seqFile);
	open (SEQ, "<$seqFile") or die "Cannot open file containing sequences, $seqFile ($!)\n";
		my ($header, $tempAcc, $tempName, $tempSeq);
		my $fastaAcc;
		my (%nexusSpecies, %nexusAcc, $nexusRead, $commentFlag);
		my ($phylipLineCount, $phylipTaxa, $phylipChars, %phylipSeq);
		my $sealCode;
		my ($sealDelFlag, $owner) = (0, 0);
		my ($gbAcc, $gbRead);
		my $macBlock = 0;
		my %accCount;

		while (<SEQ>)
			{
			chomp;
			my $lineRead = $_;
			next unless ($lineRead);
			
			# Autodetect sequence format
				if (not $inputType)
					{
					$inputType = "fasta" if ($lineRead =~ /^>/);
					$inputType = "nexus" if ($lineRead =~ /\#nexus/i);
					$inputType = "phylip" if ($lineRead =~ /^\s*\d+\s+\d+/);
					$inputType = "Se-Al" if ($lineRead =~ /^\s*Database=\{/i);
					$inputType = "GenBank" if ($lineRead =~ /^\s*LOCUS/);
					print "\nReading in sequence data from file $seqFile (type determined to be $inputType) ...\n" if ($inputType);
					}
			
			if ($inputType eq "nexus")
				{
				# Check if charset statement present
					if ($lineRead =~ /charset/i)
						{
						$lineRead =~ s/\s+//g;
						$lineRead =~ s/;$//;

						my ($charsetName, $charsetBounds) = split('=', $lineRead);
							$charsetName =~ s/charset//i;
								push @charsetList, $charsetName;
							if ($charsetBounds =~ /(\d+)-*(\d*)/)
								{
								$charsetStart{$charsetName} = $1;
								if ($2)
									{
									$charsetEnd{$charsetName} = $2;
									}
								else
									{
									$charsetEnd{$charsetName} = $1;
									}
								}
						}
						
				# Otherwise block out MacClade / PAUP blocks
					$macBlock = 1 if ($lineRead =~ /begin macclade;/i or $lineRead =~ /begin paup;/i);
					$macBlock = 0 if ($macBlock and $lineRead =~ /end;/i);
					next if ($macBlock);
						
				# Otherwise only read in data lines or charset statements
					if ($lineRead =~ /^\s*matrix/i)
						{
						$nexusRead = 1;
						next;
						}
					$nexusRead = 0 if ($lineRead =~ /;\s*$/);
					next unless ($nexusRead);
					
					# Remove MacClade sequence lengths
						$lineRead =~ s/\[\d+\]$//;

					$commentFlag = 1 if ($lineRead =~ /\[/);
					if ($lineRead =~ /\]/)
						{
						$commentFlag = 0;
						next;
						}
					next if ($commentFlag);

					next unless ($lineRead =~ /a/i or $lineRead =~ /c/i or $lineRead =~ /g/i or $lineRead =~ /t/i or $lineRead =~ /n/i or $lineRead =~ /\?/ or $lineRead =~ /\-/);

				# Clean up input line
					$lineRead =~ s/^\s+//;
#					$lineRead =~ s/\'//g;

				my @nexusLine = split(/\s+/, $lineRead);
					my $species = shift(@nexusLine);
					my $tempAcc;
						if ($species =~ /^'/)	# Contains non-alphanumerics, possibly including spaces, possibly including accession numbers
							{
							until ($species =~ /'$/)
								{
								$species .= "_" . shift(@nexusLine);
								}
							# Check for accession numbers
								if ($species =~ /(\(?\w+\d+\.?\d*\)?)/)
									{
									$tempAcc = $1;
										$tempAcc =~ s/\(//;
										$tempAcc =~ s/\)//;
									$species =~ s/$tempAcc//;
										$species =~ s/\(\)//;
										$species =~ s/^'//;
										$species =~ s/'$//;
									}
							}

						$species =~ s/\s+/_/g;
						$species =~ s/\_+/_/g;
						$species =~ s/\_$//g;
						
						$nexusAcc{$species} = $tempAcc if (defined $tempAcc);

					my $seq = join('', @nexusLine);
						$seq =~ s/\s+//g;
						$seqType = "protein" if ($seq =~ /E/i or $seq =~ /Q/i or $seq =~ /I/i or $seq =~ /L/i or $seq =~ /F/i or $seq =~ /P/i);

				if (not defined $nexusSpecies{$species})
					{
					$nexusSpecies{$species} = 1;
					$seqCount++;
					$nexusAcc{$species} = "tAlign_".$seqCount unless (defined $nexusAcc{$species});
					push @accNum, $nexusAcc{$species};
						$nameLabel{$nexusAcc{$species}} = $species;
						$sequence{$nexusAcc{$species}} = uc($seq);
						$geneticCode{$nexusAcc{$species}} = $globalGenCode;
					}
				else	# Sequences are in interleaved format; append sequence
					{
					$sequence{$nexusAcc{$species}} .= uc($seq);
					}
				}

			if ($inputType eq "fasta")
				{
				if ($lineRead =~/^\s*>/)
					{
					my $species;
					$seqCount++;
					(my $tempSpecies = $lineRead) =~ s/^\s*>//;
					
					# Parse different formats
						if ($tempSpecies =~ /^Mit\.\s+/)	# Entry comes from European RNA project
							{
							$tempSpecies =~ s/^Mit\.\s+//i;	# To fix entries from European RNA project
							my @speciesInfo = split(/\s+/, $tempSpecies);
								$species = join('_', $speciesInfo[0], $speciesInfo[1]);
							if (defined $speciesInfo[2])
								{
								$fastaAcc = $speciesInfo[2];
								$accCount{$fastaAcc}++;
									if ($accCount{$fastaAcc} > 1)
										{
										print "\nWARNING: Accession number $fastaAcc used more than once";
											print "; skipping this and subsequent entries" if ($skipDuplicates);
										print "\n";
										}
								}
							else
								{
								$fastaAcc = "tAlign_".$seqCount;
								$accCount{$fastaAcc}++;
								}
							}
						elsif ($tempSpecies =~ /^(\w+\.\d+\.\d+)/)	# Raw SILVA / ARB RNA project
							{
							$fastaAcc = $1;
							$accCount{$fastaAcc}++;
								if ($accCount{$fastaAcc} > 1)
									{
									print "\nWARNING: Accession number $fastaAcc used more than once";
										print "; skipping this and subsequent entries" if ($skipDuplicates);
									print "\n";
									}
									
							($species = $tempSpecies) =~ s/^\w+\.\d+\.\d+//;
								$species =~ s/^\s+//;
								$species =~ s/\s+/_/g;
							}
						else
							{
							# Parse Seaview specific code
								if ($tempSpecies =~ /\s+\/transl_table=(\d+)$/)
									{
									$sealCode = $1;
									$tempSpecies =~ s/\s+\/transl_table=\d+$//;
									}
								if ($tempSpecies =~ /_(\(?[A-Z]{1,2}_*\d{5,6}\.?\d*\)?)$/)
									{
									my $tempAcc = $1;
										$tempSpecies =~ s/$tempAcc//;
										$tempSpecies =~ s/\(\)//;
										$tempSpecies =~ s/_$//;
										$tempSpecies .= "\t$tempAcc";
									}
							my @speciesLine = split(/\s+/, $tempSpecies);
							if (($speciesLine[$#speciesLine] =~ /^\(?[A-Z]{1,2}_*\d{5,6}\.?\d*\)?$/ or $speciesLine[$#speciesLine] =~ /\w+\.\d+\.\d+/) and scalar(@speciesLine) > 1)	# Check whether last entry is an accession number
								{
								$fastaAcc = pop (@speciesLine);
								$fastaAcc =~ s/^\(//g;
								$fastaAcc =~ s/\)$//g;
								$accCount{$fastaAcc}++;
									if ($accCount{$fastaAcc} > 1)
										{
										print "\nWARNING: Accession number $fastaAcc used more than once";
											if ($skipDuplicates)
												{
												print "; skipping this and subsequent entries\n";
												}
											else
												{
												print "; assigning temporary accession number\n";
												$fastaAcc = "tAlign_" . $seqCount;
												}
										}
								}
							else
								{
								$fastaAcc = "tAlign_".$seqCount;
								$accCount{$fastaAcc}++;
								}
							$species = join('_', @speciesLine);
								$species = "Sequence_".$seqCount if ($species eq "");
							}
					push @accNum, $fastaAcc unless ($accCount{$fastaAcc} > 1 and $skipDuplicates);
						$geneticCode{$fastaAcc} = $globalGenCode;
					($nameLabel{$fastaAcc} = $species) =~ s/\s+/_/;
						$nameLabel{$fastaAcc} =~ s/\_+/_/;
					}
				else
					{
					next if ($accCount{$fastaAcc} > 1 and $skipDuplicates);
					$lineRead =~ s/\s+//g;
						$seqType = "protein" if ($lineRead =~ /E/i or $lineRead =~ /Q/i or $lineRead =~ /I/i or $lineRead =~ /L/i or $lineRead =~ /F/i or $lineRead =~ /P/i);
					$sequence{$fastaAcc} .= uc($lineRead);
					}
				$geneticCode{$fastaAcc} = $sealCode;
				}

			if ($inputType eq "Se-Al")
				{
				my $header;
				$sealDelFlag = 1 if ($lineRead =~ /MCoL/ or $lineRead =~ /^\}\;/);	# Se-Al sometimes places deleted species at end of file; do not read in remainder of file
					next if ($sealDelFlag == 1);
				next unless ($lineRead =~ /NumSites=/ or $lineRead =~ /Owner=/ or $lineRead =~ /Name=/ or $lineRead =~ /Accession=/ or $lineRead =~ /Sequence=/ or $lineRead =~/GeneticCode=/ or $lineRead =~ /Frame=/);
					$lineRead =~ s/^\s+//;

				if ($lineRead =~ /^Owner\s*\=\s*(\d+)/)
					{
					$owner = $1;
					}
				if ($lineRead =~ /^Accession=/ and $owner == 2)
					{
					$seqCount++;
					if ($lineRead =~ /null/ or $lineRead =~ /\"\"/)
						{
						$tempAcc = "tAlign_" . $seqCount;
						$accCount{$tempAcc}++;
						}
					else
						{
						($header, $tempAcc) = split (/=/, $lineRead);
							$tempAcc =~ s/\"//g;
							$tempAcc =~ s/;//g;
							$accCount{$tempAcc}++;
								if ($accCount{$tempAcc} > 1)
									{
									print "\nWARNING: Accession number $tempAcc used more than once";
										if ($skipDuplicates)
											{
											print "; skipping this and subsequent entries\n";
											}
										else
											{
											print "; assigning temporary accession number\n";
											$tempAcc = "tAlign_" . $seqCount;
											}
									print "\n";
									}
						}
					}
				if ($lineRead =~ /^Name=/ and $owner == 2)
					{
					($header, $tempName) = split (/=/, $lineRead);
						$tempName =~ s/\"//g;
						$tempName =~ s/\s*;//g;
						$tempName =~ s/\s+/_/g;
						$tempName =~ s/\_+/_/g;
					}
				if ($lineRead =~ /^GeneticCode=(\d+)/ and $owner == 2)
					{
					$sealCode = $1;
					$geneticCode{$tempAcc} = $seal2gb{$sealCode};
					}
				if ($lineRead =~ /^Sequence=/ and $owner == 2)
					{
					next if ($accCount{$tempAcc} > 1 and $skipDuplicates);
					($header, $tempSeq) = split (/=/, $lineRead);
						$tempSeq =~ s/\"//g;
						$tempSeq =~ s/;//g;
						$tempSeq =~ s/\s+//g;
						next if ($tempSeq eq "null");
						
					push @accNum, $tempAcc unless ($accCount{$tempAcc} > 1 and $skipDuplicates);
					$nameLabel{$tempAcc} = $tempName;
					$sequence{$tempAcc} = uc($tempSeq);
						$seqType = "protein" if ($tempSeq =~ /E/i or $tempSeq =~ /Q/i or $tempSeq =~ /I/i or $tempSeq =~ /L/i or $tempSeq =~ /F/i or $tempSeq =~ /P/i);
					}
				if ($lineRead =~ /^Frame=(\d)/)	# Correct for reading frame
					{
					my $readingFrame = $1;
					$sequence{$tempAcc} = "--" . $sequence{$tempAcc} if ($readingFrame == 2);
					$sequence{$tempAcc} = "-" . $sequence{$tempAcc} if ($readingFrame == 3);
					}
				}

			if ($inputType eq "phylip")
				{
				if ($lineRead =~ /^\s*(\d+)\s+(\d+)/)
					{
					$seqCount = $phylipTaxa = $1;
					$phylipChars = $2;
					$phylipLineCount = 0;
					}
				else
					{
					$phylipLineCount++;
					
					$lineRead =~ s/\s//g;
					
					$phylipSeq{$phylipLineCount} .= $lineRead;
					
					$phylipLineCount = 0 if ($phylipLineCount == $phylipTaxa);
					}
				}

			if ($inputType eq "GenBank")
				{
				# Get species name and accession number
					# Pure GenBank format
						$gbAcc = $1 if ($lineRead =~ /^\s*ACCESSION\s+(\w+\d+)/);
						if ($lineRead =~ /^\s*ORGANISM\s+/)
							{
							$seqCount++;
								$gbAcc = "tAlign_" . $seqCount if (not defined $gbAcc);
							$lineRead =~ s/^\s+//;
							my @orgLine = split(/\s+/, $lineRead);
								my $header = shift (@orgLine);
								$nameLabel{$gbAcc} = join('_', @orgLine);
								$accCount{$gbAcc}++;
							}
					# BioEdit format
						if ($lineRead =~ /^\s*TITLE/ and not defined ($gbAcc))
							{
							$seqCount++;
								$gbAcc = "tAlign_" . $seqCount;
							my @accLine = split (/\s+/, $lineRead);
								$gbAcc = $1 if ($accLine[2] =~ /^\((\w+\d+)\)/);
							$nameLabel{$gbAcc} = $accLine[1];
							$accCount{$gbAcc}++;
							}

				if ($lineRead =~ /^\s*ORIGIN/)
					{
					if ($accCount{$gbAcc} > 1)
						{
						print "\nWARNING: Accession number $gbAcc used more than once";
						if ($skipDuplicates)
							{
							print "; skipping this and subsequent entries\n";
							next;
							}
						else
							{
							print "; assigning temporary accession number\n";
							$gbAcc = "tAlign_" . $seqCount;
							$gbRead = 1;
							}
						}
					else
						{
						$gbRead = 1;
						$seqCount++;
						}
					next;
					}

				if ($lineRead =~ /^\s*\/\//)	# End of accession; process
					{
					$gbRead = 0;
					next unless ($gbAcc);
					
					push @accNum, $gbAcc unless ($accCount{$gbAcc} > 1);
					$geneticCode{$gbAcc} = $globalGenCode;
					$nameLabel{$gbAcc} =~ s/\s+/_/g;
						$nameLabel{$gbAcc} =~ s/\_+/_/g;
					$sequence{$gbAcc} =~ s/\d//g;
						$sequence{$gbAcc} =~ s/\s+//g;
						$sequence{$gbAcc} =~ s/\~/-/g if ($gapChange);
					undef $gbAcc;
					}
					
				next unless ($gbRead);
				
				if ($gbRead and $lineRead =~ /^\s+\d+/)
					{
					$sequence{$gbAcc} .= uc($lineRead);
					}
				}
			}
	close SEQ;
	
	if ($inputType eq "phylip")	# Postprocess input to derive taxon names and sequence; accounts for both sequential and extended formatting
		{
		for (my $i = 1; $i <= $phylipTaxa; $i++)
			{
			my $phylipAcc = "tAlign_" . $i;
			
			push @accNum, $phylipAcc;
			$geneticCode{$phylipAcc} = $globalGenCode;
			
			# Derive taxon name and sequence
				$sequence{$phylipAcc} = uc(substr($phylipSeq{$i}, 0 - $phylipChars));
					$seqType = "protein" if ($sequence{$phylipAcc} =~ /E/i or $sequence{$phylipAcc} =~ /Q/i or $sequence{$phylipAcc} =~ /I/i or $sequence{$phylipAcc} =~ /L/i or $sequence{$phylipAcc} =~ /F/i or $sequence{$phylipAcc} =~ /P/i);

				$nameLabel{$phylipAcc} = substr($phylipSeq{$i}, 0, length($phylipSeq{$i}) - $phylipChars);
					$nameLabel{$phylipAcc} =~ s/\s+/_/g;
					$nameLabel{$phylipAcc} =~ s/\_+/_/g;
			}
		}
	}

sub seqPrint
	{
	my $seqFile = shift;

	$interleaveLength = $maxLength if ($outFormat eq "sequential");
	
	# Ensure that sequence has full length
		foreach my $entry (@accNum)
			{
			next if ($deletedSeq{$entry});
			$finalSeq{$entry} .= "-" x ($maxLength - length($finalSeq{$entry}));
			}

	# Print fasta-formatted file
		if ($fastaPrint)
			{
			print "\tWriting to fasta-formatted file $fastaOut ...\n";
			open (FASTA, ">$fastaOut") or die "Cannot open fasta file for aligned DNA sequences, $fastaOut ($!)";
				foreach my $entry (@accNum)
					{
					next if ($deletedSeq{$entry});
					print FASTA ">$nameLabel{$entry}";
						print FASTA "\t($entry)" unless ($entry =~ /^tAlign/ or $entry eq $nameLabel{$entry});
						print FASTA "\n";
						
					# Print sequence
						my $fastaSeq = $finalSeq{$entry};
						for (my $breakpoint = 0; $breakpoint <= length($fastaSeq); $breakpoint += $fastaLength)
							{
							print FASTA substr($fastaSeq, $breakpoint, $fastaLength) . "\n";
							}
					}
			close FASTA;
			}

	# Print mega-formatted file
		if ($megaPrint)
			{
			print "\tWriting to mega-formatted file $megaOut ...\n";
			open (MEGA, ">$megaOut") or die "Cannot open mega file for aligned DNA sequences, $megaOut ($!)";
				print MEGA "#Mega\n";
				print MEGA "!Title $seqFile;\n\n";
				
				foreach my $entry (@accNum)
					{
					next if ($deletedSeq{$entry});
					print MEGA "#$nameLabel{$entry}\n";
					print MEGA "$finalSeq{$entry}\n\n";
					}
			close MEGA;
			}

	# Print nexus-formatted file
		if ($nexusPrint)
			{
			print "\tWriting to nexus file $nexusOut";
				print " in interleaved format" if ($outFormat eq "interleave");
				print " ...\n";
			open (NEX, ">$nexusOut") or die "Cannot open nexus file for aligned DNA sequences, $nexusOut ($!)";
				print NEX "#NEXUS\n\n";
				print NEX "[File created from $seqFile using $perlScript v$version on ".localtime()."]\n\n";
				print NEX "begin data;\n";
				print NEX "\tdimensions ntax = $ntax nchar = $maxLength;\n";
				print NEX "\tformat datatype = $seqType gap = - missing = ?";
					print NEX " interleave" if ($outFormat eq "interleave");
					print NEX ";\n\n";
				print NEX "\tmatrix\n\n";
					for (my $interleaveCount = 1; $interleaveCount <= ceil($maxLength / $interleaveLength); $interleaveCount++)
						{
						my $seqStart = (($interleaveCount - 1) * $interleaveLength) + 1;
						my $seqEnd = $interleaveCount * $interleaveLength;
							$seqEnd = $maxLength if ($maxLength <= $seqEnd);
							my $printLength = $seqEnd - $seqStart + 1;
						
						foreach my $entry (@accNum)
							{
							next if ($deletedSeq{$entry});
							my $nexusName = $nameLabel{$entry};
								$nexusName .= "_($entry)" if ($accPrint eq "append");
								$nexusName = $entry if ($accPrint eq "instead");
							if ($nexusName =~ /\W/)
								{
								print NEX "'$nexusName'";
								}
							else
								{
								print NEX "$nexusName";
								}
								
							my $printSeq = substr($finalSeq{$entry}, $seqStart - 1, $printLength);
								if ($outFormat eq "interleave")	# Add gaps every 10 elements in sequence
									{
									for (my $gapAdd = 100; $gapAdd >= 10; $gapAdd-=10)
										{
										next if $gapAdd > $printLength;
										my $bp = substr($printSeq, $gapAdd - 1, 1) . " ";
										substr($printSeq, $gapAdd - 1, 1, $bp);
										}
									}
								print NEX "\t$printSeq\n";
							}
						print NEX "\n" if ($interleaveCount * $interleaveLength <= $maxLength);
						}
				print NEX "\t;\nend;\n";
			close NEX;

			if (-e "/Developer/Tools/SetFile")
				{
				system ("/Developer/Tools/SetFile -t 'TEXT' $nexusOut");
				system ("/Developer/Tools/SetFile -c 'PAUP' $nexusOut");
				}
			}

	# Print phylip-formatted file
		if ($phylipTradPrint or $phylipExtPrint)
			{
			my $maxTaxLength = 50;
				$maxTaxLength = 10 if ($phylipTradPrint);
				my $blankName = " " x $maxTaxLength;
			my %shortNameCount;	
				
			print "\tWriting to phylip file $phylipOut ...\n";
			open (PHYLIP, ">$phylipOut") or die "Cannot open phylip file for aligned DNA sequences, $phylipOut ($!)";
				print PHYLIP "    $ntax    $maxLength\n";
				for (my $interleaveCount = 1; $interleaveCount <= ceil($maxLength / $interleaveLength); $interleaveCount++)
					{
					my $seqStart = (($interleaveCount - 1) * $interleaveLength) + 1;
					my $seqEnd = $interleaveCount * $interleaveLength;
						$seqEnd = $maxLength if ($maxLength <= $seqEnd);
						my $printLength = $seqEnd - $seqStart + 1;

					foreach my $entry (@accNum)
						{
						next if ($deletedSeq{$entry});
						
						my $phylipName = $nameLabel{$entry};
							$phylipName .= "_($entry)" if ($accPrint eq "append");
							$phylipName = $entry if ($accPrint eq "instead");
	
						# Print name label as appropriate; also check label and adjust to proper length if needed
							if ($interleaveCount == 1)
								{
								if (length($phylipName) < $maxTaxLength)
									{
									$shortNameCount{$phylipName}++;
									$phylipName .= " " x ($maxTaxLength - length($phylipName)) if ($phylipTradPrint);	# Pad end of name with spaces as needed
									}
								else
									{
									my $trimmedName = substr($phylipName, 0, $maxTaxLength);
									$shortNameCount{$trimmedName}++;
									if ($shortNameCount{$trimmedName} > 1)	# Check for duplicates among shortened names and make unique by adding numbers
										{
										$phylipName = substr($phylipName, 0, $maxTaxLength - length($shortNameCount{$trimmedName}));
											$phylipName .= $shortNameCount{$trimmedName};
											$phylipName .= " " x ($maxTaxLength - length($phylipName));	# Pad end of name with spaces as needed
										}
									else
										{
										$phylipName = $trimmedName;
										}
									}
								print PHYLIP "$phylipName";
								print PHYLIP " " if ($phylipExtPrint);
								}
							else
								{
								print PHYLIP "$blankName";
								}
								
						# Print sequence
							my $printSeq = substr($finalSeq{$entry}, $seqStart - 1, $printLength);
								if ($outFormat eq "interleave")	# Add gaps every 10 elements in sequence
									{
									for (my $gapAdd = 100; $gapAdd >= 10; $gapAdd-=10)
										{
										next if $gapAdd > $printLength;
										my $bp = substr($printSeq, $gapAdd - 1, 1) . " ";
										substr($printSeq, $gapAdd - 1, 1, $bp);
										}
									}
								print PHYLIP "$printSeq\n";
						}
					print PHYLIP "\n" if ($interleaveCount * $interleaveLength <= $maxLength);
					}
			close PHYLIP;
			}

	# Print Se-Al-formatted file
		if ($sealPrint)
			{
			print "\tWriting to Se_Al file $sealOut ...\n";
			open (SEAL, ">$sealOut") or die "Cannot open Se-Al file for aligned DNA sequences, $sealOut ($!)\n";
				print SEAL "Database={\n";
				print SEAL "\tID='MLst';\n";
				print SEAL "\tOwner=null;\n";
				print SEAL "\tName=null;\n";
				print SEAL "\tDescription=null;\n";
				print SEAL "\tFlags=0;\n";
				print SEAL "\tCount=2;\n";
				print SEAL "\t{\n\t\t{\n";
				
				print SEAL "\t\t\tID='PAli';\n";
				print SEAL "\t\t\tOwner=1;\n";
				print SEAL "\t\t\tName=\"$seqFile\";\n";
				print SEAL "\t\t\tDescription=null;\n";
				print SEAL "\t\t\tFlags=0;\n";
				print SEAL "\t\t\tNumSites=$maxLength;\n";
				print SEAL "\t\t\tType=";
					if ($seqType eq "DNA")
						{
						print SEAL "\"Nucleotide\"";
						}
					else
						{
						print SEAL "\"Amino Acid\"";
						}
					print SEAL ";\n";
				print SEAL "\t\t\tFeatures=null;\n";
				print SEAL "\t\t\tColourMode=";
					if ($seqType eq "DNA")
						{
						print SEAL "1";
						}
					else
						{
						print SEAL "2";
						}
					print SEAL ";\n";
				print SEAL "\t\t\tLabelMode=0;\n";
				print SEAL "\t\t\ttriplets=false;\n";
				print SEAL "\t\t\tinverse=true;\n";
				print SEAL "\t\t\tCount=$ntax;\n";
				print SEAL "\t\t\t{\n";
				
				my $i = 0;
				foreach my $sequence (@accNum)
					{
					next if ($deletedSeq{$sequence});
					$i++;
					print SEAL "\t\t\t\t{\n";
					print SEAL "\t\t\t\t\tID='PSeq';\n";
					print SEAL "\t\t\t\t\tOwner=2;\n";
					print SEAL "\t\t\t\t\tName=\"$nameLabel{$sequence}\";\n";
					print SEAL "\t\t\t\t\tDescription=null;\n";
					print SEAL "\t\t\t\t\tFlags=0;\n";
					print SEAL "\t\t\t\t\tAccession=";
						if ($sequence =~/^tAlign_/)
							{
							print SEAL "null;\n";
							}
						else
							{
							print SEAL "$sequence;\n";
							}
					if ($seqType eq "DNA")
						{
						print SEAL "\t\t\t\t\tType=\"DNA\";\n";
						}
					else
						{
						print SEAL "\t\t\t\t\tType=\"AA\";\n";
						}
					print SEAL "\t\t\t\t\tLength=".length($finalSeq{$sequence}).";\n";
					print SEAL "\t\t\t\t\tSequence=\"$finalSeq{$sequence}\";\n";
					if (defined $geneticCode{$sequence} and $geneticCode{$sequence} != 0)
						{
						print SEAL "\t\t\t\t\tGeneticCode=$gb2seal{$geneticCode{$sequence}};\n";
						}
					else
						{
						print SEAL "\t\t\t\t\tGeneticCode=-1;\n";	# Default for Se-Al is non-coding
						}
					print SEAL "\t\t\t\t\tCodeTable=null;\n";
					print SEAL "\t\t\t\t\tFrame=1;\n";
					print SEAL "\t\t\t\t\tFeatures=null;\n";
					print SEAL "\t\t\t\t\tParent=null;\n";
					print SEAL "\t\t\t\t\tComplemented=false;\n";
					print SEAL "\t\t\t\t\tReversed=false;\n";
					print SEAL "\t\t\t\t}";
					print SEAL "," unless ($i == $ntax);
					print SEAL "\n";
					}
				
				print SEAL "\t\t\t};\n";
				print SEAL "\t\t},\n";
				print SEAL "\t\t{\n";
				print SEAL "\t\t\tID='MCoL';\n";
				print SEAL "\t\t\tOwner=1;\n";
				print SEAL "\t\t\tName=\"Genetic Codes\";\n";
				print SEAL "\t\t\tDescription=\"Custom Genetic Codes\";\n";
				print SEAL "\t\t\tFlags=0;\n";
				print SEAL "\t\t\tCount=0;\n";
				print SEAL "\t\t}\n";
				print SEAL "\t};\n";
				print SEAL "};\n";
			close SEAL;

			if (-e "/Developer/Tools/SetFile")
				{
				system ("/Developer/Tools/SetFile -t 'TEXT' $sealOut");
				system ("/Developer/Tools/SetFile -c 'SEAL' $sealOut");
				}
			}

	# Print stockholm-formatted file
		if ($stockholmPrint)
			{
			print "\tWriting to stockholm-formatted file $stockholmOut ...\n";
			open (STOCKHOLM, ">$stockholmOut") or die "Cannot open stockholm file for aligned DNA sequences, $stockholmOut ($!)";
				print STOCKHOLM "# STOCKHOLM 1.0\n";
				foreach my $entry (@accNum)
					{
					next if ($deletedSeq{$entry});
					print STOCKHOLM "$nameLabel{$entry}\t$finalSeq{$entry}\n";
					}
				print STOCKHOLM "//\n";
			close STOCKHOLM;
			}
	}

sub matrixRead
	{
	my $inFile = shift;

	my $prevLine = "";
	my ($taxonRead, $matrixRead) = (0, 0);
		my ($tstatusFlag, $matrixFlag) = (0 ,0);
	my %taxonRawName;
	
	# Variables to guard against PAUP* line wrapping
		my $wrapFlag = 0;
		my $numColumns;
		my $lastTaxon = 0;
		my $matrixLine;
	
	setLineBreak($inFile);
	open (IN, "<$inFile") or die "Cannot open $inFile to read matrix ($!).\n";
		while (<IN>)
			{
			chomp;
			print "$_\n" if ($debug);
			
			$paupVersion = $1 if ($_ =~ /version 4.0(\w+)/i);
			
			# Set reading flags
				if ($_ =~ /distance matrix$/ or $_ =~ /^No data for comparing taxa/)
					{
					$taxonRead = 0;
					$matrixRead = 1;
					$matrixFlag = 1;
					}

				if ($_ =~ /Current status of all taxa/)
					{
					$taxonRead = 1;
					$matrixRead = 0;
					$tstatusFlag = 1;
					print "Reading in taxon status ...\n" if ($debug);
					}
					
				next if ($_ =~ /^No data for comparing taxa/);
					
			# Read in taxon IDs; accounts for spaces in taxon names
				if ($taxonRead)
					{
					my $nameLength = 16;	# PAUP* only prints out first 16 or 30 characters in distance matrix
						$nameLength = 30 if ($paupVersion =~/a/);
						
					next unless ($_ =~ /\s?\d+/);
						$_ =~ s/^\s+//;
					
					my @taxonLine = split(/\s+/, $_);
						my $taxonID = shift(@taxonLine);
						pop(@taxonLine);	# Get rid of last entry indicating taxon status
						$taxonRawName{$taxonID} = join(' ', @taxonLine);
							$taxonRawName{$taxonID} = substr($taxonRawName{$taxonID}, 0, $nameLength);	# PAUP* only prints out first 16 or 30 characters in distance matrix
						$taxonName{$taxonID} = join('_', @taxonLine);
						print "\t$taxonID = $taxonName{$taxonID} ($taxonRawName{$taxonID})\n" if ($debug);
					}

			# Read in matrix
				if ($matrixRead)
					{
					die "ERROR: Failed to read in taxon status information (PAUP command = tstatus full).\n" unless ($tstatusFlag);
					# Skip lines not related to matrix
						$_ =~ s/^\s+//;
#							$_ =~ s/\*//g;
							next if ($_ =~ /^[a-zA-Z]/);
							next if ($_ =~ /^\*/);
							next if ($_ =~ /characters are included/);

 					# Mark blank lines, which indicate new portion of matrix coming			
						if (not $_)
							{
							$prevLine = "blank";
							next;
							}
							
					# Read in matrix
						my $readLine = $_;
							my @readLine = split(/\s+/, $readLine);
						
						if ($prevLine eq "blank")	# Read in header line
							{
							$startColumn = shift(@readLine);
							$endColumn = pop(@readLine);
								$endColumn = $startColumn if (not defined $endColumn);	# Only one column in matrix
								$numTaxa = $endColumn;
							$numColumns = $endColumn - $startColumn + 1;
							$prevLine = "";
							next;
							}
						else	# Process matrix
							{
							if ($debug)
								{
								print "Read: $readLine\n";
								print "\ttaxonID: $readLine[0]" if (defined $readLine[0]);
								print "\ttaxonName: $taxonName{$readLine[0]}" if (defined $readLine[1]);
								print "\ttaxonRawName: $taxonRawName{$readLine[0]}" if (defined $taxonRawName{$readLine[0]});
								print "\n\n";
								}
							# Safety to check if PAUP* has wrapped part of matrix on to a new line
								if (defined $readLine[0] and defined $taxonRawName{$readLine[0]} and $taxonRawName{$readLine[0]} =~ /^$readLine[1]/)	# Check that start of line formats as taxonID<white space>taxonName
									{
									$rowTaxon = $readLine[0];
										$readLine =~ s/$taxonRawName{$rowTaxon}/$taxonName{$rowTaxon}/;	# Correct line for taxon names with spaces
									$matrixLine = $readLine;
										print "\tNormal line: $readLine ($startColumn)\n" if ($debug);
									}
								else	# Wrapped line; append to previous line and reparse
									{
									$matrixLine .= " $readLine";
									print "\t\tMatrix: $matrixLine ($startColumn)\n" if ($debug);
									}

							@readLine = split(/\s+/, $matrixLine);

							my $columnTaxon = $startColumn - 1;
							for (my $column = 2; $column < scalar(@readLine); $column++)
								{
								$columnTaxon++;
								my $value = $readLine[$column];
									$columnTaxon = $rowTaxon if ($value eq "-");
									print "$rowTaxon\t$columnTaxon\t$value\n" if ($debug);
#print "$rowTaxon\t$columnTaxon\t$value\n";
#print "$taxonName{$rowTaxon}\t$taxonName{$columnTaxon}\t$value\n";
									if ($value =~ /\*$/)
										{
										next if ($ignoreUndefined);
										$value =~ s/\*$//;
										}
								
								if ($columnTaxon > $rowTaxon)
									{
									$aboveDiag{$rowTaxon}{$columnTaxon} = $aboveDiag{$columnTaxon}{$rowTaxon} = $value;
									$aboveDiag{$taxonName{$rowTaxon}} {$taxonName{$columnTaxon}} = $aboveDiag{$taxonName{$columnTaxon}} {$taxonName{$rowTaxon}} = $value;
									$aboveFlag = 1;
									}
								if ($columnTaxon < $rowTaxon)
									{
									$belowDiag{$rowTaxon}{$columnTaxon} = $value;
									$belowDiag{$columnTaxon}{$rowTaxon} = $value;
									$belowDiag{$taxonName{$rowTaxon}} {$taxonName{$columnTaxon}} = $belowDiag{$taxonName{$columnTaxon}} {$taxonName{$rowTaxon}} = $value;
									$belowFlag = 1;
									}
								}
							}
					}
			}
	# Error checking
		die "ERROR: Failed to read in pairwise distance matrix.\n" unless ($matrixFlag);

	return;
	}

sub parseTree
	{
	my $parseID = shift;
		my $parseTree = $treeDescription{$parseID};

	print "Parsing structure ...\n" unless ($quiet);

	# Do general check on tree structure
		# Check tree size
			my $totalClade = ($parseTree =~ tr/(//);
				if ($totalClade < 2 and $bushCheck)
					{
					print "\tERROR: Tree has only one informative node; skipping\n";
					return;
					}

		# Check for branch lengths
			my $branchFlag = ($parseTree =~ m/:\d+/) ? 1 : 0;
				$branchType = "none" unless ($branchFlag);

	# Clear variables
		undef %ancCladeHeight;
		undef %dauCladeHeight;
		
		undef @nodeList;
		undef %ParentClade;
			undef %daughterClade;
		undef %CladeDescrip;
			undef %open;
			undef %start;
			undef %stop;
		undef %geneNodeMember;
			undef %taxonClade;
			undef @GTtaxa;
			undef %cladeSize;
		undef %CladeSupport;
			undef %nodeDepth;
			undef %cladeDepth;
			undef %cladeHeight;
			# Set variables for root
				$nodeDepth{GTnode0} = 0;
				$cladeDepth{GTnode0} = 0;
				$CladeSupport{GTnode1} = 0;	# Default; will be overriden by any real value		
		undef $taxonFlag;
		my $taxonRead = 0;
		$parseError = 0;

		my ($position, $cladeNum, $CurrentClade) = (0, 0, 0);
		my $OldCurrent;
		my $taxon;
		my $cladeID;
	
	while ($position < length($parseTree))
		{
		my $offset = 1;
		if (substr($parseTree, $position, 1) eq "(")		# New clade; go one level deeper
			{
			$cladeNum++;
				$cladeID = "GTnode" . $cladeNum;
				push @nodeList, $cladeID;
			$ParentClade{$cladeID} = $CurrentClade;
				$ParentClade{$cladeID} = "GTnode0" if ($cladeID eq "GTnode1");
			push @ {$daughterClade{$CurrentClade} }, $cladeID;	# Assign new clade as daughter clade to parent clade
			$CurrentClade = $cladeID;
				$cladeSize{$CurrentClade} = 0;
			$nodeDepth{$CurrentClade} = $nodeDepth{$ParentClade{$CurrentClade}} + 1;
			$open{$CurrentClade} = 1;
			$start{$cladeID} = $position;
			}
		elsif (substr($parseTree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
			{
			$open{$CurrentClade} = 0;
			$OldCurrent = $CurrentClade;
			$CurrentClade = $ParentClade{$OldCurrent};
			$stop{$OldCurrent} = $position;
			$taxonFlag = 0;
			# Get clade description
				$CladeDescrip{$OldCurrent} = substr($parseTree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
				$CladeDescrip{$OldCurrent} =~ s/:\-?\d+\.?\d*e?\-?\d*//ig if ($removeBL);	# Remove branch length information
				
			$polyClade{$OldCurrent} = (scalar(@ {$daughterClade{$OldCurrent} }) > 2) ? 1 : 0;

			print "\t\t\tClade ID: $OldCurrent\n" if ($debug);
			print "\t\t\t\t$CladeDescrip{$OldCurrent}\n" if ($debug);
			
			# Determine absolute and relative heights from tips
				if ($branchFlag)
					{
					my $sumHeight = 0;
					print "\t\t\t\tDaughter clades:\n" if ($debug);
					unless ($branchType eq "support")
						{
						foreach my $daughter (@ {$daughterClade{$OldCurrent} })
							{
							print "\t\t\t\t\t$daughter: $CladeSupport{$daughter}\n" if ($debug);
							$sumHeight += $CladeSupport{$daughter};							
							$sumHeight += $cladeHeight{$daughter};							
							}
						print "\t\t\t\t\tSum height: $sumHeight\n" if ($debug);
						$cladeHeight{$OldCurrent} = sprintf "%.5f", $sumHeight / scalar(@ {$daughterClade{$OldCurrent} });
							print "\t\t\t\t\tClade height: $cladeHeight{$OldCurrent}\n" if ($debug);
						}
					}
			}
		elsif (substr($parseTree, $position, 1) eq ":")	# Store branch lengths for node or taxon
			{
			until (substr($parseTree, $position+$offset, 1) eq "(" or substr($parseTree, $position+$offset, 1) eq ")" or substr($parseTree, $position+$offset, 1) eq "," or ($position+$offset) >= length($parseTree))
				{
				$offset++;
				}
			my $branchLength = substr($parseTree, $position+1, $offset-1);
				print "\t\t\t\tBranch length: $branchLength\n" if ($debug);
				if ($taxonFlag == 0)	# Branch length pertains to a clade
					{
					$CladeSupport{$OldCurrent} = substr($parseTree, $position+1, $offset-1);
					}
				else	# Branch length pertains to a terminal taxon
					{
					$CladeSupport{$taxon} = $branchLength;
					$CladeDescrip{$taxon} .= ":$branchLength" unless ($removeBL);
					}
			$position += $offset-1;
			}
		elsif (substr($parseTree, $position, 1) eq ",")
			{
			# Do nothing
			}
		else	# Determine membership of clade
			{
			my $prevChar = substr($parseTree, $position-1, 1);
			$taxonRead = 1 if (substr($parseTree, $position, 1) eq "'");

			until ($position+$offset >= length($parseTree) or (not $taxonRead and (substr($parseTree, $position+$offset, 1) eq "(" or substr($parseTree, $position+$offset, 1) eq ")" or substr($parseTree, $position+$offset, 1) eq "," or substr($parseTree, $position+$offset, 1) eq ":")))
				{
				$offset++;
				if (substr($parseTree, $position+$offset-1, 1) eq "'")
					{
					$taxonRead = ($taxonRead) ? 0 : 1;
					}
				}
			$taxon = substr($parseTree, $position, $offset);

			if ($prevChar eq ")")	# Text represents a node label and not a taxon
				{
				$taxon =~ s/^\'//;
				$taxon =~ s/\'$//;
				$nodeLabel{$OldCurrent} = $taxon;
				$labelFlag = 1;
				}
			else
				{
				push @ { $daughterClade{$CurrentClade}}, $taxon;
				push @GTtaxa, $taxon;
				push @nodeList, $taxon;
				$ParentClade{$taxon} = $CurrentClade;
				$nodeDepth{$taxon} = $nodeDepth{$CurrentClade} + 1;
				$cladeHeight{$taxon} = 0;
				$cladeSize{$taxon} = 1;
				$CladeDescrip{$taxon} = $taxon;
				print "\t\t\tTaxon: $taxon\n" if ($debug);
				$taxonFlag = 1;

				# Assign taxon to any open clades
					for (my $i = 1; $i <= $cladeNum; $i++)
						{
						my $clade = "GTnode" . $i;
						if ($open{$clade})
							{
							push @{ $geneNodeMember{$clade} }, $taxon;
							$taxonClade{$clade}{$taxon} = 1;
							$cladeSize{$clade}++;
							}
						else
							{
							$taxonClade{$clade}{$taxon} = 0; 
							}
						}
				}
			$position += $offset-1;
			}
		$position++;
		}
		
	$parseError = 1 if ($CurrentClade ne "GTnode0");
	@GTtaxa = sort(@GTtaxa);
	}

sub BootStrip
	{
	# BootStrip.pl v1.3
	# Last modified September 8, 2011 14:05
	# (c) Olaf R.P. Bininda-Emonds
	#
	# Input:
	#   1) A treefile containing one or more tree descriptions in NEXUS or phylip
	#	   formats 
	#   2) Either: a) a set of bootstrap trees in NEXUS or phylip formats
	#			   b) a PAUP* logfile of a bootstrap analysis containing a table of
	#				  group frequencies
	#
	# Generates:
	#   A nexus-formatted treefile (BootStrip.tre) containing the trees in (1)
	#   with bootstrap frequencies derived from (2) represented as branch lengths,
	#	node labels, or both
	#
	# Usage: BootStrip.pl [-a] [-b<filename>] [-d<number>] [-f<l|n|b>] [-i<number>]
	#		 [-k] [-l<filename>] [-p] [-pp] [-r<rootword>] [-t<filename>] [-v] [-h]
	#   options: -a = print out average bootstrap frequency and SE across all
	#				  informative nodes in tree
	#			 -b<filename> = file containing bootstrap tree in NEXUS or phylip
	#                           formats (default = bootTrees.tre)
	#			 -d<number> = number of decimal places to print bootstrap
	#						  frequencies (default = 3)
	#            -f<l|n|b> = include bootstrap frequencies as branch lengths (l;
	#						 default), node labels (n), or both (b)
	#			 -i<number> = only process trees beyond this replicate number (for
	#						  MrBayes output in conjunction with -b and not -p)
	#			 -k = keep branch-length information in tree (will override -f for
	#				  that tree)
	#            -l<filename> = log file of bootstrap analysis (default =
	#							bootstrap.log)
	#			 -p = use PAUP* to summarize bootstrap trees (faster than built-in
	#				  mechanism) (default name = paup in \$PATH)
	#			 -pp = summarize support as probabilities (i.e., 1.00 rather than
	#				   100)
	#			 -r<rootword> = common rootword for -b (rootword_bootTrees.tre),
	#							-l (rootword_bootlog.txt), and -t (rootword.tre)
	#							flags (as appropriate)
	#            -t<filename> = file containing optimal tree(s) in NEXUS or phylip
	#							formats (default = treeFile.tre)
	#            -v = verbose output
	#            -h = print this message and quit
	
	my $bootSwitches = shift;
		my @bootSwitches = split(/\s+/, $bootSwitches);
		my $runID = shift(@bootSwitches);
		undef %bootFreq;
		undef %plotCount;

	# Process switches
		for (my $i = 0; $i <= $#bootSwitches; $i++)
			{
			if ($bootSwitches[$i] eq "-a")
				{
				$averagePrint = 1;
				}
			elsif ($bootSwitches[$i] =~ /^-b(.+)/)
				{
				$bootFile = $1;
				die "ERROR: Bootstrap treefile $bootFile does not exist.\n" unless (-e $bootFile);
				undef $logFile;
				}
			elsif ($bootSwitches[$i] =~ /^-d([\d]+)/)
				{
				$decimalPlace = $1;
				}
			elsif ($bootSwitches[$i] eq "-fl")
				{
				$labelType = "branch";
				}
			elsif ($bootSwitches[$i] eq "-fn")
				{
				$labelType = "node";
				}
			elsif ($bootSwitches[$i] eq "-fb")
				{
				$labelType = "both";
				}
			elsif ($bootSwitches[$i] =~ /^-i([\d]+)/)
				{
				$burninLimit = $1;
				}
			elsif ($bootSwitches[$i] eq "-k")
				{
				$keepLengths = 1;
				}
			elsif ($bootSwitches[$i] =~ /^-l(.+)/)
				{
				$logFile = $1;
				die "ERROR: Bootstrap logfile $logFile does not exist.\n" unless (-e $logFile);
				undef $bootFile;
				}
			elsif ($bootSwitches[$i] eq "-p")
				{
				$paupFlag = 1;
				}
			elsif ($bootSwitches[$i] eq "-pp")
				{
				$postProb = $tipValue = 1;
				}
			elsif ($bootSwitches[$i] =~ /^-r/)
				{
				($rootWord = $bootSwitches[$i]) =~ s/^-r//;
				$optFile = $rootWord . ".tre";
					die "ERROR: Treefile $optFile does not exist.\n" unless (-e $optFile);
				$logFile = $rootWord . "_bootlog.txt";		
				$bootFile = $rootWord . "_bootTrees.tre";		
					die "ERROR: Logfile $logFile and bootstrap tree file $bootFile do not exist.\n" unless (-e $logFile or -e $bootFile);
	
				$outFile = $rootWord . "_bootstrip.tre";
				}
			elsif ($bootSwitches[$i] =~ /^-t(.+)/)
				{
				$optFile = $1;
				die "ERROR: Treefile $optFile does not exist.\n" unless (-e $optFile);
#				($outFile = $optFile) =~ s/\.\w+$//;
#				($outFile = $optFile) =~ s/$runID.+$/$runID/;
				($outFile = $optFile) =~ s/\.RUN\.\d+$//;
				$outFile .= "_bootstrip.tre";
				}
			elsif ($bootSwitches[$i] eq "-v")
				{
				$verbose = 1;
				}
			elsif ($bootSwitches[$i] eq "-x")
				{
				$debug = $verbose = 1;
				}
			elsif ($bootSwitches[$i] eq "-h")
				{
				print "Usage: BootStrip.pl [-a] [-b<filename>] [-d<number>] [-f<l|n|b>] [-i<number>]\n";
				print "       [-k] [-l<filename>] [-p] [-pp] [-r<rootword>] [-t<filename>] [-v] [-h]\n";
				print "Version: $version\n";
				print "Options: -a = print out average bootstrap frequency and SE across all\n";
				print "              informative nodes in tree\n";
				print "         -b<filename> = file containing bootstrap tree in NEXUS or phylip\n";
				print "                        formats (default = bootTrees.tre)\n";
				print "         -d<number> = number of decimal places to print bootstrap\n";
				print "                      frequencies (default = 3)\n";
				print "         -f<l|n|b> = include bootstrap frequencies as branch lengths (l;\n";
				print "                     default), node labels (n), or both (b)\n";
				print "         -i<number> = only process trees beyond this replicate number (for\n";
				print "                      MrBayes output in conjunction with -b and not -p)\n";
				print "         -k = keep branch-length information in tree (will override -f for\n";
				print "              that tree)\n";
				print "         -l<filename> = log file of bootstrap analysis (default =\n";
				print "                        bootstrap.log)\n";
				print "         -p = use PAUP* to summarize bootstrap trees (faster than built-in\n";
				print "              mechanism) (default name = paup in \$PATH)\n";
				print "         -pp = summarize support as probabilities (i.e., 1.00 rather than\n";
				print "               100)\n";
				print "         -r<rootword> = common rootword for -b (rootword_bootTrees.tre),\n";
				print "                        -l (rootword_bootlog.txt) and -t (rootword.tre)\n";
				print "                        flags (as appropriate)\n";
				print "         -t<filename> = file containing optimal tree(s) in NEXUS or phylip\n";
				print "                        formats (default = treeFile.tre)\n";
				print "         -v = verbose output\n";
				print "         -h = print this message and quit\n";
				exit(0);
				}
			else
				{
				print "Don't understand argument: $ARGV[$i]\n";
				print "Usage: BootStrip.pl [-a] [-b<filename>] [-d<number>] [-f<l|n|b>] [-i<number>]\n";
				print "       [-k] [-l<filename>] [-p] [-pp] [-r<rootword>] [-t<filename>] [-v] [-h]\n";
				print "Version: $version\n";
				exit(1); 
				}
			}
			
		my $decimalPrint = "%.$decimalPlace"."f";
	
	# Read in optimal trees
		print "\nReading in optimal tree descriptions from $optFile ...\n" if ($verbose);
			my @optTrees = treeRead($optFile);
			die "No trees present in $optFile" if (not @optTrees);
			my @taxonList = @allTaxa;
	
	# Process bootstrap trees
		if (defined $bootFile and -e $bootFile)
			{
			# Read in and store trees
				print "Reading in bootstrap trees from file $bootFile ...\n" if ($verbose);
				my @bootTrees = treeRead($bootFile);
					$nullDot = "." x scalar(@taxonList);
					$fullDot = "*" x scalar(@taxonList);
	
			if ($paupFlag)	# Derive logfile from bootstrap trees using PAUP*
				{
				my $nexusFile = $rootWord . "_majRuler.nex";
					$logFile = "bootstrip_majRuler.txt";
				my ($startTree, $endTree);
				open (NEX, ">$nexusFile") or die "Cannot write to nexus-formatted treefile, $nexusFile\n";
					print NEX "#NEXUS\n\n";
					print NEX "begin taxa;\n";
						printf NEX "\tdimensions ntax = %s;\n", scalar(@taxonList);
						print NEX "\ttaxlabels\n";
						print NEX "\t\t$_\n" foreach (@taxonList);
						print NEX "\t\t;\n";
						print NEX "end;\n\n";
						
					if (scalar(@bootTrees) > 100)	# Add paup block to override initial maxtrees setting of 100
						{
						print NEX "begin paup;\n";
							printf NEX "\tset maxtrees=%s increase=no;\n", scalar(@bootTrees);
						print NEX "end;\n\n";
						}
			
					print NEX "begin trees;\n";
						print NEX "\ttree $treeName{$_} = $rootingStatus{$_} $treeDescription{$_};\n" foreach (@bootTrees);
						print NEX "end;\n";
			
					print NEX "\nbegin paup;\n";
						print NEX "\tset taxlabels=full outroot=monophyl;\n";
						print NEX "\tcontree ";
							print NEX "$startTree - $endTree " if (defined $startTree and defined $endTree);
							print NEX "/ strict=no majrule=yes le50=yes grpfreq=yes replace;\n";
						print NEX "end;\n";
						
					print  NEX "quit / warntsave=no;\n";
				close NEX;
			
				if (-e "/Developer/Tools/SetFile")
					{
					system ("/Developer/Tools/SetFile -t 'TEXT' $nexusFile");
					system ("/Developer/Tools/SetFile -c 'PAUP' $nexusFile");
					}
			
				system ("$paup $nexusFile > $logFile");
				}
			else
				{
				# Read in and store trees
					print "Reading in bootstrap trees from file $bootFile ...\n" if ($verbose);
					my @bootTrees = treeRead($bootFile);
					
				# Process taxon list and assign IDs
					for (my $i = 0; $i < scalar(@taxonList); $i++)
						{
						$taxonID{$taxonList[$i]} = $i + 1;
						}
			
				# Obtain frequencies of dot plots across bootstrap trees
					my (@rawBootPlots, @bootPlots);
					my $treeCount = 0;
			
					foreach my $treeID (@bootTrees)
						{
						my $repNumber = $1 if ($treeName{$treeID} =~ /(\d+)/);
							next if ($repNumber <= $burninLimit);
						push @rawBootPlots, dotPlotter($rootingStatus{$treeID}, $treeDescription{$treeID});
						$treeCount++;
						}
	
					@bootPlots = duplStrip(@rawBootPlots);
						foreach my $plot (@bootPlots)
							{
							$bootFreq{$plot} = sprintf ("$decimalPrint", $plotCount{$plot} / $treeCount * 100);
								$bootFreq{$plot} /= 100 if ($postProb);
							}
						$minFreq = sprintf ("$decimalPrint", 1 / $treeCount * 100);
							$minFreq /= 100 if ($postProb);
				}
			}
	
	# Process logfile as needed to read in bootstrap frequencies and taxon IDs
		if (defined $logFile and -e $logFile)
			{
			setLineBreak($logFile);
			open (LOG, "<$logFile") or die "Cannot open bootstrap logfile $logFile ($!)\n";
				my (%paupID, %dotPlot, @dotPlot, $numReps);
				my $bootReadFlag = 0;
				my $dotReadFlag = 0;
				my $lineCount = 0;
				my $numBipartitions = 0;
			
				while (<LOG>)
					{
					chomp;
					my $treeLine = $_;
						$treeLine =~ s/^\s+//;
						$treeLine =~ s/\s+$//;
					next unless ($treeLine);
						
					# Get number of bootstrap replicates
						if ($treeLine =~ /Number of bootstrap replicates/)
							{
							(my $header, $numReps) = split(/=/, $treeLine);
							$numReps =~ s/\s+//g;
							}
						if ($treeLine =~ /^50% Majority-rule consensus of (\d+) tree/)
							{
							$numReps = $1;
							}
		
					# Set flags to only read in appropriate lines
						$bootReadFlag = 1 if ($treeLine =~ /Bootstrap 50%/i or $treeLine =~ /^50% Majority-rule consensus/i);
						if ($treeLine =~ /\(relative\) frequency less than ([\d]+)/i)
							{
							$minFreq = $1;
								$minFreq /= 100 if ($postProb);
							$bootReadFlag = $dotReadFlag = 0;
							}
						if ($treeLine =~ /^Processing of file/)
							{
							$minFreq = 0.1 if (not defined $minFreq);
								$minFreq /= 100 if ($postProb);
							$bootReadFlag = $dotReadFlag = 0;
							}
						next unless ($bootReadFlag);
		
					# Read in taxon IDs	
						if ($treeLine =~ /\(\d+\)/)
							{
							my ($header, $fullTaxon) = split(/\-+\s+/, $treeLine);
							my ($name, $ID) = split(/\(/, $fullTaxon);
								$name =~ s/\s+/_/g;
								$ID =~ s/\)\s*$//;
							$paupID{$name} = $ID;			
							}
					
					# Read in group frequency table
						$dotReadFlag = 1 if ($treeLine =~ /\(bootstrap support values\)\:/i or $treeLine =~ /^Bipartitions found in one or more trees/i or $treeLine =~ /^Groups found in one or more trees and frequency of occurrence/i);
						
						$lineCount = 0 if ($treeLine =~ /^-------/ and $dotReadFlag);
							$lineCount++ if ($dotReadFlag and ($treeLine =~ /^\./ or $treeLine =~ /^\*/));
							$numBipartitions++ if ($lineCount > $numBipartitions);
							
						$dotPlot{$lineCount} .= $treeLine if ($dotReadFlag and ($treeLine =~ /^\./ or $treeLine =~ /^\*/));
					}
			close LOG;
			
			die "No taxon IDs present in $logFile" if (not %paupID);
			die "No group frequencies present in $logFile" if (not $numBipartitions);
			die "ERROR: Could not determine number of replicates.\n" unless ($numReps);
			
			# Process dot plots to derive bootstrap frequencies
				for (my $plotNum = 1; $plotNum <= $numBipartitions; $plotNum++)
					{
		#			print "$plotNum: $dotPlot{$plotNum}\n" if ($debug);
					my (@plotLine) = split(/\s+/, $dotPlot{$plotNum});	# plotLine columns are dotPlot, raw frequency, and (majRule only) percentage
						print "$dotPlot{$plotNum}\n" if ($debug);
		
					push @dotPlot, $plotLine[0];
					$bootFreq{$plotLine[0]} = sprintf("$decimalPrint", $plotLine[1] / $numReps * 100);
						$bootFreq{$plotLine[0]} /= 100 if ($postProb);
					}
		
		# Get taxon list and process
			@taxonList = sort { $paupID{$a} <=> $paupID{$b} } keys %paupID;
				$nullDot = "." x scalar(@taxonList);
				$fullDot = "*" x scalar(@taxonList);
			
			# Renumber paup IDs for all taxa to remove any gaps (due to taxon deletion)
				for (my $i = 0; $i < scalar(@taxonList); $i++)
					{
					$taxonID{$taxonList[$i]} = $i + 1;
					}
			}
	
	# Process optimal trees
		print "\nParsing tree descriptions and assigning bootstrap frequencies ...\n" if ($verbose);
	
		my $treeCount = 0;
		open (OUT, ">$outFile") or die "Cannot open output file $outFile\n";
			print OUT "#NEXUS\n\nbegin trees;\n";
			foreach my $treeID (@optTrees)	# Get structure of each tree
				{
				$treeCount++;
					print "\nTree: $treeCount\n" if ($verbose);
					
				my ($sumBoot, $avgBoot, $varBoot, $numBoot) = (0, 0, 0, 0);
	
				my $tree = $treeDescription{$treeID};
					my $lengthFlag = 0;
						$lengthFlag = 1 if ($tree =~ /:\d+\.?\d*e?\-?\d*/i);	# Check if tree has branch-length information
					my $treeGrow;
	
				my ($offset, @ParentClade, @open, @start, @stop, $OldCurrent, @CladeDescrip);
				my ($position, $CladeID, $CurrentClade) = (0, 0, 0);
				my %present;
				while ($position < length($tree))
					{
					$offset = 1;
					if (substr($tree, $position, 1) eq "(")		# New clade; go one level deeper
						{
						$CladeID++;
						$ParentClade[$CladeID] = $CurrentClade;
						$CurrentClade = $CladeID;
						$open[$CurrentClade] = 1;
						$start[$CladeID] = $position;
						$treeGrow .= "(";
						}
					elsif (substr($tree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
						{
						$open[$CurrentClade] = 0;
						$OldCurrent = $CurrentClade;
						$CurrentClade = $ParentClade[$OldCurrent];
						$stop[$OldCurrent] = $position;
						$CladeDescrip[$OldCurrent] = substr($tree, $start[$OldCurrent], $stop[$OldCurrent]-$start[$OldCurrent]+1);
						$CladeDescrip[$OldCurrent] =~ s/:\-?\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
						my $presentDot = $nullDot;
						my $absentDot = $fullDot;
						print "\t$OldCurrent\t$CladeDescrip[$OldCurrent]\n" if ($verbose);
						if ($CurrentClade == 0)
							{
							$treeGrow .= ")";
							$treeGrow .= "'n/a'" if ($labelType eq "both" or $labelType eq "node");
							$treeGrow .= ":0" if ($labelType eq "both" or $labelType eq "branch");
							print "\t\tBootFreq: n/a\n" if ($verbose);
							}
						else
							{
							$treeGrow .= ")";
							my $cladeBF;
							foreach my $taxon (@taxonList)	# Derive dot plot for that clade
								{
								if ($present{$taxon}{$OldCurrent})
									{
									substr($presentDot, $taxonID{$taxon}-1, 1) = "*";
									substr($absentDot, $taxonID{$taxon}-1, 1) = ".";
									}
								}
							print "\t\t$presentDot\n\t\t$absentDot\n" if ($verbose);
							
							# Check that bipartition is meaningful (i.e., smallest clade has 2+ members)
								my $bipartCount = ($presentDot =~ tr/\*//);
								if ($bipartCount == 1 or length($presentDot) - $bipartCount == 1)
									{
									if ($rootingStatus{$treeID} eq "[&U]")
										{
										$treeGrow .= "'n/a'" if ($labelType eq "both" or $labelType eq "node");
										if ($labelType eq "both" or $labelType eq "branch")
											{
											$treeGrow .= ":0" unless ($keepLengths and $lengthFlag);
											}
										print "\t\tBootFreq: n/a\n" if ($verbose);
										}
									else
										{
										$treeGrow .= "'".(sprintf "$decimalPrint", $tipValue)."'" if ($labelType eq "both" or $labelType eq "node");
										if ($labelType eq "both" or $labelType eq "branch")
											{
											$treeGrow .= ":".(sprintf "$decimalPrint", $tipValue) unless ($keepLengths and $lengthFlag);
											}
										print "\t\tBootFreq: n/a\n" if ($verbose);
										}
									}
								else
									{
									if (defined $bootFreq{$presentDot} or defined $bootFreq{$absentDot})
										{
										my $printFreq;
											$printFreq = $bootFreq{$presentDot} if (defined $bootFreq{$presentDot});
											$printFreq = $bootFreq{$absentDot} if (defined $bootFreq{$absentDot});
										
										$treeGrow .= "'".(sprintf "$decimalPrint", $printFreq)."'" if ($labelType eq "both" or $labelType eq "node");
										if ($labelType eq "both" or $labelType eq "branch")
											{
											$treeGrow .= ":".(sprintf "$decimalPrint", $printFreq) unless ($keepLengths and $lengthFlag);
											}
										$numBoot++;
										$sumBoot += $printFreq;
										$varBoot += $printFreq * $printFreq;
										}
									else	# No match: give minimal frequency
										{
										print "\t\tBootFreq: <$minFreq\n" if ($verbose);
										$treeGrow .= "'<$minFreq'" if ($labelType eq "both" or $labelType eq "node");
										if ($labelType eq "both" or $labelType eq "branch")
											{
											$treeGrow .= ":".(sprintf "$decimalPrint", $minFreq) unless ($keepLengths and $lengthFlag);
											}
										$numBoot++;
										$sumBoot += $minFreq;
										$varBoot += $minFreq * $minFreq;
										}
									}
							}
						}
					elsif (substr($tree, $position, 1) eq ",")
						{
						$treeGrow .= ",";
						}
					elsif (substr($tree, $position, 1) eq ":")	# Branch lengths present; ignore
						{
						until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or $position+$offset >= length($tree))
							{
							$offset++;
							}
						$treeGrow .= substr($tree, $position, $offset) if ($keepLengths == 1 and $lengthFlag == 1);
						$position += $offset-1;
						}
					else	#Determine membership of clade
						{
						my $prevChar = substr($tree, $position-1, 1);
						until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or substr($tree, $position+$offset, 1) eq ":" or $position+$offset >= length($tree))
							{
							$offset++;
							}
						unless ($prevChar eq ")")	# Text represents a node label and not a taxon
							{
							for (my $clade = 1; $clade <= $CladeID; $clade++)	#Assign taxon to any open clades
								{
								if ($open[$clade])
									{
									$present{substr($tree, $position, $offset)}{$clade} = 1;
									}
								}
							$treeGrow .= substr($tree, $position, $offset);
							if ($labelType eq "both" or $labelType eq "branch")
								{
								$treeGrow .= ":" . sprintf "$decimalPrint", $tipValue unless ($keepLengths == 0 and $lengthFlag == 0);
								}
							}
						$position += $offset-1;
						}
					$position++;
					}
					$treeGrow =~ s/,+/,/g;
					$treeGrow =~ s/,\)/\)/g;
			
				print OUT "\ttree $treeName{$treeID} = $rootingStatus{$treeID} $treeGrow;\n";
					print "\ttree $treeName{$treeID} = $rootingStatus{$treeID} $treeGrow;\n"if ($verbose);
					if ($averagePrint)
						{
						if ($numBoot > 1)
							{
							$avgBoot = $sumBoot / $numBoot;
							$varBoot = ($varBoot - ($sumBoot * $sumBoot) / $numBoot) / ($numBoot - 1);
							
							printf OUT "\t\t[Average bootstrap frequency: $decimalPrint +/- $decimalPrint (n = $numBoot)]\n", $avgBoot, sqrt ($varBoot / $numBoot);
								$varBoot = 0 if ($varBoot < 0);
							}
						elsif ($numBoot == 1)
							{
							printf OUT "\t\t[Average bootstrap frequency: $decimalPrint (SE undefined; n = $numBoot)\n", $avgBoot;
							}
						else
							{
							printf OUT "\t\t[Average bootstrap frequency: undefined (n = $numBoot)\n";
							}
						}
				}
			print OUT "end;\n";
			close OUT;
		
	return;
	}

sub dotPlotter
	{
	my $rootingStatus = shift;
		print "Rooting status: $rootingStatus\n" if ($verbose);
	my $tree = shift;
		print "Tree description:\n\t$tree\n" if ($verbose);
	
	my (%ParentClade, %CladeDescrip, %CladeSupport, %open, %start, %stop, %present);
	my (@taxonList, @dotPlots, %taxonID, %dotPlot, %treePlotCount);
	
	my $totalClade = ($tree =~ tr/(//);
	my ($position, $CladeID, $CurrentClade) = (0, 0, 0);
	my $OldCurrent;
	
	# Get list of taxa and assign IDs
		(my $taxonString = $tree) =~ s/:\-?\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
			$taxonString =~ s/\(//g;
			$taxonString =~ s/\)//g;
		@taxonList = sort(split(/,/, $taxonString));
			for (my $i = 0; $i < scalar(@taxonList); $i++)
				{
				$taxonID{$taxonList[$i]} = $i;
				}

	# Parse tree description and derive dot plots		
		while ($position < length($tree))
			{
			my $offset = 1;
			if (substr($tree, $position, 1) eq "(")		# New clade; go one level deeper
				{
				$CladeID++;
				$ParentClade{$CladeID} = $CurrentClade;
				$CurrentClade = $CladeID;
				$open{$CurrentClade} = 1;
				$start{$CladeID} = $position;
				$dotPlot{$CladeID} = "." x scalar(@taxonList);	# Create blank dot plot for clade
				}
			elsif (substr($tree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
				{
				$open{$CurrentClade} = 0;
				$OldCurrent = $CurrentClade;
				$CurrentClade = $ParentClade{$OldCurrent};
				$stop{$OldCurrent} = $position;
				$CladeDescrip{$OldCurrent} = substr($tree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
				$CladeDescrip{$OldCurrent} =~ s/:\-?\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
				}
			elsif (substr($tree, $position, 1) eq ":")	# Branch lengths present; record as nodal support
				{
				until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or $position+$offset >= length($tree))
					{
					$offset++;
					}
				$position += $offset-1;
				}
			elsif (substr($tree, $position, 1) eq ",")
				{
				# Do nothing
				}
			elsif (substr($tree, $position, 1) eq "[")	# Skip over embedded comment
				{
				until (substr($tree, $position+$offset, 1) eq "]" or $position+$offset >= length($tree))
					{
					$offset++;
					}
				$position += $offset-1;
				}
			else	# Determine membership of clade
				{
				my $prevChar = substr($tree, $position-1, 1);
				until (substr($tree, $position+$offset, 1) eq "(" or substr($tree, $position+$offset, 1) eq ")" or substr($tree, $position+$offset, 1) eq "," or substr($tree, $position+$offset, 1) eq ":" or $position+$offset >= length($tree))
					{
					$offset++;
					}
				unless ($prevChar eq ")")	# Text represents a node label and not a taxon
					{
					my $taxon = substr($tree, $position, $offset);
					for (my $clade = 1; $clade <= $CladeID; $clade++)	# Assign taxon to any open clades
						{
						if ($open{$clade})
							{
							$present{substr($tree, $position, $offset)}{$clade} = 1;
							substr($dotPlot{$clade}, $taxonID{$taxon}, 1) = "*";
							}
						}
					}
				$position += $offset-1;
				}
			$position++;
			}

	# Check that dot plots are informative and invert as needed
		for (my $clade = 1; $clade <= $totalClade; $clade++)
			{
			print "Clade $clade:\n\t$CladeDescrip{$clade}\n" if ($verbose);
			my $dotPlot = $dotPlot{$clade};

			# Only store informative clades
				next unless ($dotPlot =~ /\./ and $dotPlot =~ /\*/);
				if ($rootingStatus eq "unrooted" or $rootingStatus eq "[&U]")
					{
					my $starCount = ($dotPlot =~ tr/\*//);
					next if ($starCount == 1 or $starCount == (length($dotPlot) - 1));
					}

			$dotPlot = invertPlot($dotPlot) if ($dotPlot =~ /^\*/ and ($rootingStatus eq "unrooted" or $rootingStatus eq "[&U]"));	# Invert plot to common frame of reference to account for unrooted trees

			$treePlotCount{$dotPlot}++;
				print "\tDot plot: $dotPlot\n" if ($verbose);
				if ($treePlotCount{$dotPlot} == 1)	# Only store unique plots (and unique to that tree)
					{
					push @dotPlots, $dotPlot;
					$plotCount{$dotPlot}++;
					}
				else
					{
					print "\t\tNot unique ...\n" if ($verbose);
					}
			}
			
	return @dotPlots;
	}

sub invertPlot	# Change *s to .s (and vice versa) in dot plot
	{
	my $plot = shift;
		$plot =~ s/\./x/g;
		$plot =~ s/\*/\./g;
		$plot =~ s/x/\*/g;
		
	return $plot;
	}

# Version history
#
#	v1.0 (October 3, 2016)
#		- initial release
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is available at
# http://www.gnu.org/copyleft/gpl.html or by writing to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Fifth Floor, Boston, MA, 02110-1301, USA.
